from collections import Counter
from pathlib import Path

import pandas as pd
import plotly.express as px
from pandas import DataFrame

from immuneML.data_model import bnp_util, AIRRSequenceSet
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset, SequenceDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class SequenceLengthDistribution(DataReport):
    """
    Generates a histogram of the lengths of the sequences in a dataset.


    **Specification arguments:**

    - sequence_type (str): whether to check the length of amino acid or nucleotide sequences; default value is 'amino_acid'

    - region_type (str): which part of the sequence to examine; e.g., IMGT_CDR3

    - split_by_label (bool): Whether to split the plots by a label. If set to true, the Dataset must either contain a
      single label, or alternatively the label of interest can be specified under 'label'. By default,
      split_by_label is False.

    - label (str): if split_by_label is set to True, a label can be specified here.

    - plot_frequencies (bool): if set to True, the plot will show the frequencies of the sequence lengths instead of the
      counts. By default, plot_frequencies is False.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_sld_report:
                    SequenceLengthDistribution:
                        sequence_type: amino_acid
                        region_type: IMGT_CDR3
                        label: label_1
                        split_by_label: True
                        plot_frequencies: True

    """

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_sequence_type(kwargs)
        ParameterValidator.assert_region_type(kwargs)

        return SequenceLengthDistribution(**{**kwargs, 'sequence_type': SequenceType[kwargs['sequence_type'].upper()],
                                             'region_type': RegionType[kwargs['region_type'].upper()]})

    def __init__(self, dataset: Dataset = None, batch_size: int = 1, result_path: Path = None,
                 number_of_processes: int = 1, region_type: RegionType = RegionType.IMGT_CDR3,
                 sequence_type: SequenceType = SequenceType.AMINO_ACID, name: str = None, label: str = None,
                 split_by_label: bool = False, plot_frequencies: bool = False):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.batch_size = batch_size
        self.sequence_type = sequence_type
        self.region_type = region_type
        self.label_name = label
        self.split_by_label = split_by_label
        self.plot_frequencies = plot_frequencies

    def check_prerequisites(self):
        return True

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        self.label_name = self._get_label_name()

        df = self._get_sequence_lengths_df()
        df.to_csv(self.result_path / 'sequence_length_distribution.csv', index=False)

        report_output_fig = self._safe_plot(df=df, output_written=False)
        output_figures = None if report_output_fig is None else [report_output_fig]
        return ReportResult(name=self.name,
                            info="A histogram of the lengths of the sequences in a dataset.",
                            output_figures=output_figures,
                            output_tables=[ReportOutput(self.result_path / 'sequence_length_distribution.csv',
                                                        'lengths of sequences in the dataset')])

    def _get_sequence_lengths_df(self) -> DataFrame:
        if isinstance(self.dataset, RepertoireDataset):
            sequence_lengths_df = self._get_sequence_lengths_df_repertoire_dataset()
        elif isinstance(self.dataset, SequenceDataset):
            sequence_lengths_df = self._get_sequence_lengths_df_sequence_dataset()
        elif isinstance(self.dataset, ReceptorDataset):
            sequence_lengths_df = self._get_sequence_lengths_df_receptor_dataset()

        if not self.split_by_label and self.label_name in sequence_lengths_df.columns:
            sequence_lengths_df.drop(columns=[self.label_name], inplace=True)

        if self.plot_frequencies:
            if self.split_by_label and 'chain' not in sequence_lengths_df.columns:
                sequence_lengths_df['frequencies'] = sequence_lengths_df.groupby(self.label_name)['counts'].transform(
                    lambda x: x / x.sum())
            elif self.split_by_label and 'chain' in sequence_lengths_df.columns:
                sequence_lengths_df['frequencies'] = sequence_lengths_df.groupby([self.label_name, 'chain'])[
                    'counts'].transform(lambda x: x / x.sum())
            else:
                sequence_lengths_df['frequencies'] = sequence_lengths_df['counts'] / sequence_lengths_df['counts'].sum()

        return sequence_lengths_df

    def _get_sequence_lengths_df_repertoire_dataset(self):
        raw_count_dict_per_class = {}
        label_name = self._get_label_name()

        if self.split_by_label:
            class_names = self.dataset.get_metadata([label_name])[label_name]
        else:
            class_names = [0] * self.dataset.get_example_count()

        for repertoire, class_name in zip(self.dataset.get_data(), class_names):
            self._count_seq_lengths(repertoire.data, raw_count_dict_per_class, class_name)

        return self._count_dict_per_class_to_df(raw_count_dict_per_class)

    def _get_sequence_lengths_df_sequence_dataset(self):
        return self._count_dict_per_class_to_df(self._count_seq_lengths(self.dataset.data, class_name=None))

    def _count_dict_per_class_to_df(self, raw_count_dict_per_class):
        result_dfs = []

        for class_name, raw_count_dict in raw_count_dict_per_class.items():
            result_df = self._count_dict_to_df(raw_count_dict)
            result_df[self.label_name] = class_name
            result_dfs.append(result_df)

        return pd.concat(result_dfs)

    def _count_seq_lengths(self, data: AIRRSequenceSet, raw_count_dict=None, class_name: str = None):
        raw_count_dict = {} if raw_count_dict is None else raw_count_dict

        for item in data.to_iter():
            sequence = getattr(item, bnp_util.get_sequence_field_name(self.region_type, SequenceType.AMINO_ACID))

            if self.split_by_label and self.label_name is not None:
                if class_name is None:
                    cls_name = getattr(item, self.label_name)
                else:
                    cls_name = class_name
            else:
                cls_name = 0

            if cls_name not in raw_count_dict:
                raw_count_dict[cls_name] = {}

            if len(sequence) not in raw_count_dict[cls_name]:
                raw_count_dict[cls_name][len(sequence)] = 1
            else:
                raw_count_dict[cls_name][len(sequence)] += 1

        return raw_count_dict

    def _count_dict_to_df(self, count_dict):
        return pd.DataFrame({"counts": list(count_dict.values()),
                             'sequence_lengths': list(count_dict.keys())})

    def _get_dataset_chains(self):
        return next(self.dataset.get_data()).get_chains()

    def _get_sequence_lengths_df_receptor_dataset(self):
        data = self.dataset.data
        chains = list(set(data.locus.tolist()))

        dfs = []
        label_name = self._get_label_name()

        for chain in chains:
            chain_data = data[[el == chain for el in data.locus.tolist()]]

            if self.split_by_label and label_name:
                chain_df = self._count_dict_per_class_to_df(self._count_seq_lengths(chain_data, class_name=None))
                chain_df["chain"] = chain
                dfs.append(chain_df)
            else:
                chain_counter = Counter(getattr(chain_data,
                                                bnp_util.get_sequence_field_name(self.region_type,
                                                                                 self.sequence_type)).lengths.tolist())
                dfs.append(pd.DataFrame({'counts': list(chain_counter.values()),
                                         'sequence_lengths': list(chain_counter.keys()),
                                         'chain': chain}))

        return pd.concat(dfs)

    def _count_in_repertoire(self, repertoire: Repertoire) -> Counter:
        return Counter(getattr(repertoire.data,
                               bnp_util.get_sequence_field_name(self.region_type,
                                                                self.sequence_type)).lengths.tolist())

    def _plot(self, df: pd.DataFrame) -> ReportOutput:

        figure = px.bar(df, x="sequence_lengths", y="frequencies" if self.plot_frequencies else "counts",
                        facet_col=self.label_name if self.label_name in df.columns else None,
                        facet_row="chain" if isinstance(self.dataset, ReceptorDataset) else None)
        figure.update_layout(template="plotly_white")
        figure.update_traces(marker_color=px.colors.qualitative.Vivid[1])

        for annotation in figure.layout.annotations:
            annotation['font'] = {'size': 16}

        PathBuilder.build(self.result_path)

        file_path = self.result_path / "sequence_length_distribution.html"
        file_path = PlotlyUtil.write_image_to_file(figure, file_path)
        return ReportOutput(path=file_path, name="Sequence length distribution plot")

    def _get_label_name(self):
        if self.split_by_label:
            if self.label_name is None:
                return list(self.dataset.get_label_names())[0]
            else:
                return self.label_name
        else:
            return None



import abc
from pathlib import Path

from immuneML.reports.Report import Report
from immuneML.reports.ReportResult import ReportResult
from immuneML.workflows.instructions.clustering.ClusteringState import ClusteringState


class ClusteringReport(Report):

    DOCS_TITLE = "Clustering Instruction Reports"

    def __init__(self, name: str = None, result_path: Path = None, number_of_processes: int = 1,
                 state: ClusteringState = None):
        super().__init__(name, result_path, number_of_processes)
        self.state = state

    @classmethod
    def build_object(cls, **kwargs):
        pass

    @abc.abstractmethod
    def _generate(self) -> ReportResult:
        pass



from pathlib import Path
from typing import List

import pandas as pd
import plotly.express as px

from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.multi_dataset_reports.MultiDatasetReport import MultiDatasetReport
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.SequenceAnalysisHelper import SequenceAnalysisHelper


class DiseaseAssociatedSequenceOverlap(MultiDatasetReport):
    """
    DiseaseAssociatedSequenceOverlap report makes a heatmap showing the overlap of disease-associated sequences (or k-mers)
    produced by the :py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder`,
    :py:obj:`~immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder.CompAIRRSequenceAbundanceEncoder` or
    :py:obj:`~immuneML.encodings.abundance_encoding.KmerAbundanceEncoder.KmerAbundanceEncoder`
    between multiple datasets of different sizes (different number of repertoires per dataset).

    This plot can be used only with MultiDatasetBenchmarkTool.

    The overlap is computed by the following equation:

    .. math::

        overlap(X,Y) = \\frac{|X \\cap Y|}{min(|X|, |Y|)} * 100

    For details, see:
    Greiff V, Menzel U, Miho E, et al. Systems Analysis Reveals High Genetic and Antigen-Driven Predetermination of Antibody
    Repertoires throughout B Cell Development. Cell Reports. 2017;19(7):1467-1478. doi:10.1016/j.celrep.2017.04.054.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_overlap_report: DiseaseAssociatedSequenceOverlap # report has no parameters

    """

    @classmethod
    def build_object(cls, **kwargs):
        return DiseaseAssociatedSequenceOverlap(**kwargs)

    def __init__(self, instruction_states: List[TrainMLModelState] = None, name: str = None, result_path: Path = None, number_of_processes: int = 1):
        super().__init__(instruction_states=instruction_states, name=name, result_path=result_path, number_of_processes=number_of_processes)
        self.label = None

    def _generate(self) -> ReportResult:
        self.result_path = PathBuilder.build(self.result_path / self.name)
        self._extract_label()

        hp_items = [state.optimal_hp_items[self.label.name] for state in self.instruction_states]
        overlap_matrix = SequenceAnalysisHelper.compute_overlap_matrix(hp_items)

        labels = [state.dataset.name for state in self.instruction_states]
        figure_path = self._make_figure(overlap_matrix, labels)
        data_path = self._export_matrix(overlap_matrix, labels)

        return ReportResult(name=self.name,
                            info="A heatmap showing the overlap of disease-associated sequences produced by SequenceAbundance encoders between multiple datasets of different sizes.",
                            output_figures=[ReportOutput(figure_path, 'sequence overlap across datasets')],
                            output_tables=[ReportOutput(data_path, 'sequence overlap across datasets (csv)')])

    def _extract_label(self):
        all_labels = []
        for state in self.instruction_states:
            all_labels += state.label_configuration.get_label_objects()

        label_names = set([label.name for label in all_labels])
        assert len(label_names) == 1, \
            f"{DiseaseAssociatedSequenceOverlap.__name__}: multiple labels were specified {label_names}, but this report accepts only one label."

        self.label = list(all_labels)[0]

    def _export_matrix(self, overlap_matrix, labels):
        data_path = self.result_path / "sequence_overlap.csv"
        pd.DataFrame(overlap_matrix, columns=labels, index=labels).to_csv(data_path)
        return data_path

    def _make_figure(self, overlap_matrix, labels):
        figure = px.imshow(overlap_matrix, x=labels, y=labels, zmin=0, zmax=100, color_continuous_scale=px.colors.sequential.Teal,
                           template='plotly_white')
        figure.update_traces(hovertemplate="Overlap of disease-associated<br>sequences between datasets<br>%{x} and %{y}:<br>%{z}%<extra></extra>")
        figure_path = self.result_path / "sequence_overlap.html"
        figure_path = PlotlyUtil.write_image_to_file(figure, figure_path)
        return figure_path




import json
import logging
from enum import Enum

import numpy as np

from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.PathBuilder import PathBuilder


class NumpyHelper:

    SIMPLE_TYPES = [str, int, float, bool, np.str_, np.int_, np.float_, np.bool_]

    @staticmethod
    def group_structured_array_by(data, field):
        for col in data.dtype.names:
            data[col][np.argwhere(data[col] == None)] = ""
        sorted_data = np.sort(data, order=[field], axis=0)
        grouped_lists = np.split(sorted_data, np.cumsum(np.unique(sorted_data[field], return_counts=True)[1])[:-1])
        return grouped_lists

    @staticmethod
    def is_simple_type(t):
        """returns if the type t is string or a number so that it does not use pickle if serialized"""
        return t in NumpyHelper.SIMPLE_TYPES

    @staticmethod
    def get_numpy_representation(obj):
        """converts object to representation that can be stored without pickle enables in numpy arrays; if it is an object or a dict,
        it will be serialized to a json string"""

        if obj is None:
            return ''
        elif type(obj) in NumpyHelper.SIMPLE_TYPES:
            return obj
        elif isinstance(obj, Enum):
            return obj.name
        elif not isinstance(obj, dict) and not isinstance(obj, list) and not isinstance(obj, Enum):
            representation = vars(obj)
        else:
            representation = obj

        return json.dumps(representation, default=lambda x: x.name if isinstance(x, Enum) else str(x))

    @staticmethod
    def is_nan_or_empty(value):
        return value == 'nan' or value is None or (not isinstance(value, str) and np.isnan(value)) or value == ''

    @staticmethod
    def create_memmap_array_in_cache(shape: tuple, data: np.ndarray = None) -> np.ndarray:
        """Creates a memory-mapped array and optionally initializes it with data."""
        import uuid
        dir_path = PathBuilder.build(EnvironmentSettings.get_cache_path() / "memmap_storage")
        memmap_path = dir_path / f"temp_{uuid.uuid4()}.npy"
        if data is not None:
            data.astype('float32').tofile(memmap_path)
            return np.memmap(memmap_path, dtype='float32', mode='r+', shape=shape)
        else:
            return np.memmap(memmap_path, dtype='float32', mode='w+', shape=shape)

    @staticmethod
    def concat_arrays_rowwise(arrays: list, force='auto', dense_max_mb=100, use_memmap=False):
        """
        Concatenate 2D numpy arrays or sparse matrices row-wise.

        Parameters
        ----------
        arrays : list of np.ndarray or scipy.sparse matrices
        force : {"auto", "dense", "sparse"}
            - "auto": use memory-based heuristic (default)
            - "dense": always return numpy.ndarray
            - "sparse": always return scipy.sparse.csr_matrix
        dense_max_mb : int
            Threshold for converting sparse -> dense in "auto" mode.
        use_memmap: bool
        """
        if not arrays:
            raise ValueError("No matrices provided")

        from scipy import sparse
        if any(sparse.issparse(array) for array in arrays):
            # Convert all to sparse (CSR for efficiency)
            matrices = [array.astype(np.float32) if sparse.issparse(array)
                        else sparse.csr_matrix(array).astype(np.float32) for array in arrays]
            result = sparse.hstack(matrices, format="csr")

            # Estimate dense memory size
            size_in_mb = result.shape[0] * result.shape[1] * result.dtype.itemsize / (1024 * 1024)
            if size_in_mb <= dense_max_mb:
                result = result.toarray()
        else:
            # All are numpy arrays
            result = np.hstack(arrays)

        if np.isnan(result).any():
            import inspect
            logging.error(f"NumpyHelper: NaN values found in concatenated array; called from {inspect.stack()[1].function}")
            raise RuntimeError('NumpyHelper: NaN values found in concatenated array')

        if force == "dense" and sparse.issparse(result):
            result = result.toarray()
        elif force == "sparse" and not sparse.issparse(result):
            result = sparse.csr_matrix(result)

        if use_memmap:
            result = NumpyHelper.create_memmap_array_in_cache(result.shape, result)

        return result




import logging

import numpy as np

from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.protein_embedding.ProteinEmbeddingEncoder import ProteinEmbeddingEncoder
from immuneML.util.Logger import log_memory_usage
from immuneML.util.NumpyHelper import NumpyHelper
from immuneML.util.ParameterValidator import ParameterValidator


class ProtT5Encoder(ProteinEmbeddingEncoder):
    """
    Encoder based on a pretrained protein language model by Elnaggar et al. 2021. The used transformer model is
    "Rostlab/prot_t5_xl_half_uniref50-enc".

    Original publication:
    Elnaggar, A., Heinzinger, M., Dallago, C., Rihawi, G., Wang, Y., Jones, L., Gibbs, T., Feher, T.,
    Angerer, C., Steinegger, M., Bhowmik, D., & Rost, B. (2021). ProtTrans: Towards Cracking the Language of
    Life's Code Through Self-Supervised Deep Learning and High Performance Computing (No. arXiv:2007.06225).
    arXiv. https://doi.org/10.48550/arXiv.2007.06225

    Original GitHub repository with license information: https://github.com/agemagician/ProtTrans

    **Dataset type:**

    - SequenceDatasets

    - ReceptorDatasets

    - RepertoireDatasets

    **Specification arguments:**

    - region_type (RegionType): Which part of the receptor sequence to encode. Defaults to IMGT_CDR3.

    - device (str): Which device to use for model inference - 'cpu', 'cuda', 'mps' - as defined by pytorch.
      Defaults to 'cpu'.

    - num_processes (int): Number of processes to use for parallel processing. Defaults to 1.

    - batch_size (int): The number of sequences to encode at the same time. This could have large impact on memory usage.
      If memory is an issue, try with smaller batch sizes. Defaults to 4096.

    - scale_to_zero_mean (bool): Whether to scale the embeddings to zero mean. Defaults to True.

    - scale_to_unit_variance (bool): Whether to scale the embeddings to unit variance. Defaults to True.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_prot_t5_encoder:
                    ProtT5::
                        region_type: IMGT_CDR3
                        device: cpu
                        num_processes: 1
                        batch_size: 4096

    """

    def __init__(self, name: str = None, region_type: RegionType = RegionType.IMGT_CDR3, device: str = 'cpu',
                 num_processes: int = 1, batch_size: int = 4096, scale_to_zero_mean: bool = True,
                 scale_to_unit_variance: bool = True):
        super().__init__(region_type, name, num_processes, device, batch_size, scale_to_zero_mean=scale_to_zero_mean,
                         scale_to_unit_variance=scale_to_unit_variance)
        self.transformer_link = "Rostlab/prot_t5_xl_half_uniref50-enc"
        self.batch_size = batch_size
        self.embedding_dim = 1024  # ProtT5's output dimension
        self.mem_map_path = None

    @staticmethod
    def build_object(dataset: Dataset, **params):
        ParameterValidator.assert_region_type(params, ProtT5Encoder.__name__)
        return ProtT5Encoder(**{**params, 'region_type': RegionType[params['region_type'].upper()]})

    def _get_model_and_tokenizer(self, log_location):
        import torch
        from transformers import T5Tokenizer, T5EncoderModel

        log_memory_usage(stage="start", location=log_location)
        logging.info(f"ProtT5 ({self.name}): Loading: {self.transformer_link}")
        model = T5EncoderModel.from_pretrained(self.transformer_link)
        log_memory_usage("after model load", log_location)

        if self.device == torch.device("cpu"):
            logging.info(f"{log_location}: Casting model to full precision for running on CPU ...")
            model.to(torch.float32)

        model = model.to(self.device).eval()
        log_memory_usage("after model to device", log_location)

        tokenizer = T5Tokenizer.from_pretrained(self.transformer_link, do_lower_case=False, legacy=True)
        log_memory_usage("after tokenizer load", log_location)

        return model, tokenizer

    def _embed_sequence_set(self, sequence_set: AIRRSequenceSet, seq_field: str):
        import torch

        log_location = f"ProtT5Encoder ({self.name})"
        model, tokenizer = self._get_model_and_tokenizer(log_location)

        sequences = getattr(sequence_set, seq_field)
        sequence_lengths = sequences.lengths
        sequences = [" ".join(list(sequence)) for sequence in sequences.tolist()]
        n_sequences = len(sequences)

        # Create memory-mapped array for embeddings
        embeddings = NumpyHelper.create_memmap_array_in_cache((n_sequences, self.embedding_dim))

        for i in range(0, n_sequences, self.batch_size):
            batch_end = min(i + self.batch_size, n_sequences)
            batch = sequences[i:batch_end]
            batch_lengths = sequence_lengths[i:batch_end]

            logging.info(
                f"{log_location}: Processing batch {i // self.batch_size + 1}/{(n_sequences - 1) // self.batch_size + 1}")

            ids = tokenizer.batch_encode_plus(batch, add_special_tokens=True, padding="longest")
            input_ids = torch.tensor(ids['input_ids']).to(self.device)
            attention_mask = torch.tensor(ids['attention_mask']).to(self.device)

            with torch.no_grad():
                embedding_repr = model(input_ids=input_ids, attention_mask=attention_mask)

            batch_embeddings = [embedding_repr.last_hidden_state[j, :batch_lengths[j]].mean(dim=0).cpu().numpy()
                                for j in range(len(batch))]
            embeddings[i:batch_end] = batch_embeddings

            del embedding_repr, input_ids, attention_mask
            if torch.cuda.is_available():
                torch.cuda.empty_cache()

            log_memory_usage(f"after batch {i // self.batch_size + 1}", log_location)

        logging.info(f"{log_location}: Finished processing all sequences")
        return embeddings

    def _get_encoding_name(self) -> str:
        return f"ProtT5Encoder({self.transformer_link})"

    def _get_model_link(self) -> str:
        return self.transformer_link


from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod, get_data_for_clustering
from sklearn.cluster import HDBSCAN as SklearnHDBSCAN


class HDBSCAN(ClusteringMethod):
    """
    HDBSCAN method which wraps scikit-learn's clustering of the same name.
    Input arguments for the method are the same as supported by scikit-learn (see `DBSCAN scikit-learn documentation
    <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.HDBSCAN.html>`_ for details).

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_hdbscan:
                    HDBSCAN:
                        # arguments as defined by scikit-learn
                        min_cluster_size: 5

        """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.model = SklearnHDBSCAN(**kwargs)

    def fit(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        self.model.fit(data)

    def fit_predict(self, dataset: Dataset):
        data = get_data_for_clustering(dataset)
        return self.model.fit_predict(data)

    def predict(self, dataset: Dataset):
        raise RuntimeError("HDBSCAN Clustering does not support predict method")

    def transform(self, dataset: Dataset):
        raise RuntimeError("HDBSCAN Clustering does not support transform method")



from collections import Counter
from pathlib import Path

import logging
import pandas as pd

from immuneML.IO.ml_method.MLExporter import MLExporter
from immuneML.data_model.SequenceParams import RegionType
from immuneML.environment.Label import Label
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.example_weighting.ExampleWeightingStrategy import ExampleWeightingStrategy
from immuneML.hyperparameter_optimization.config.ReportConfig import ReportConfig
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.hyperparameter_optimization.core.HPAssessment import HPAssessment
from immuneML.hyperparameter_optimization.core.HPUtil import HPUtil
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.hyperparameter_optimization.strategy.HPOptimizationStrategy import HPOptimizationStrategy
from immuneML.ml_metrics.ClassificationMetric import ClassificationMetric
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.Logger import print_log
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.MLProcess import MLProcess
from scripts.specification_util import update_docs_per_mapping


class TrainMLModelInstruction(Instruction):
    """
    Class implementing hyperparameter optimization and training and assessing the model through nested cross-validation (CV).
    The process is defined by two loops:

    - the outer loop over defined splits of the dataset for performance assessment

    - the inner loop over defined hyperparameter space and with cross-validation or train & validation split
      to choose the best hyperparameters.

    Optimal model chosen by the inner loop is then retrained on the whole training dataset in the outer loop.

    .. note::

        If you are interested in plotting the performance of all combinations of encodings and ML methods on the test set,
        consider running the :ref:`MLSettingsPerformance` report as hyperparameter report in the assessment loop.


    **Specification arguments:**

    - dataset (Dataset): dataset to use for training and assessing the classifier

    - hp_strategy (HPOptimizationStrategy): how to search different hyperparameters; common options include grid search, random search. Valid values are objects of any class inheriting :py:obj:`~immuneML.hyperparameter_optimization.strategy.HPOptimizationStrategy.HPOptimizationStrategy`.

    - hp_settings (list): a list of combinations of `preprocessing_sequence`, `encoding` and `ml_method`. `preprocessing_sequence` is optional, while `encoding` and `ml_method` are mandatory. These three options (and their parameters) can be optimized over, choosing the highest performing combination.

    - assessment (SplitConfig): description of the outer loop (for assessment) of nested cross-validation. It describes how to split the data, how many splits to make, what percentage to use for training and what reports to execute on those splits. See plitConfig below.

    - selection (SplitConfig): description of the inner loop (for selection) of nested cross-validation. The same as assessment argument, just to be executed in the inner loop. See plitConfig below.

    - metrics (list): a list of metrics to compute for all splits and settings created during the nested cross-validation. These metrics will be computed only for reporting purposes. For choosing the optimal setting, `optimization_metric` will be used.

    - optimization_metric (Metric): a metric to use for optimization and assessment in the nested cross-validation.

    - example_weighting: which example weighting strategy to use. Example weighting can be used to up-weight or down-weight the importance of each example in the dataset. These weights will be applied when computing (optimization) metrics, and are used by some encoders and ML methods.

    - label_configuration (LabelConfiguration): a list of labels for which to train the classifiers. The goal of the nested CV is to find the
      setting which will have best performance in predicting the given label (e.g., if a subject has experienced an immune event or not).
      Performance and optimal settings will be reported for each label separately. If a label is binary, instead of specifying only its name, one
      should explicitly set the name of the positive class as well under parameter `positive_class`. If positive class is not set, one of the label
      classes will be assumed to be positive.

    - number_of_processes (int): how many processes should be created at once to speed up the analysis. For personal machines, 4 or 8 is usually a good choice.

    - reports (list): a list of report names to be executed after the nested CV has finished to show the overall performance or some statistic;
      the reports to be specified here have to be :py:obj:`~immuneML.reports.train_ml_model_reports.TrainMLModelReport.TrainMLModelReport` reports.

    - refit_optimal_model (bool): if the final combination of preprocessing-encoding-ML model should be refitted on the full dataset thus providing
      the final model to be exported from instruction; alternatively, train combination from one of the assessment folds will be used

    - export_all_models (bool): if set to True, all trained models in the assessment split are exported as .zip files.
      If False, only the optimal model is exported. By default, export_all_models is False.

    - sequence_type (str): whether to perform the analysis on amino acid or nucleotide sequences

    - region_type (str): which part of the sequence to analyze, e.g., IMGT_CDR3


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_nested_cv_instruction: # user-defined name of the instruction
                type: TrainMLModel # which instruction should be executed
                settings: # a list of combinations of preprocessing, encoding and ml_method to optimize over
                    - preprocessing: seq1 # preprocessing is optional
                      encoding: e1 # mandatory field
                      ml_method: simpleLR # mandatory field
                    - preprocessing: seq1 # the second combination
                      encoding: e2
                      ml_method: simpleLR
                assessment: # outer loop of nested CV
                    split_strategy: random # perform Monte Carlo CV (randomly split the data into train and test)
                    split_count: 1 # how many train/test datasets to generate
                    training_percentage: 0.7 # what percentage of the original data should be used for the training set
                    reports: # reports to execute on training/test datasets, encoded datasets and trained ML methods
                        data_splits: # list of reports to execute on training/test datasets (before they are encoded)
                            - rep1
                        encoding: # list of reports to execute on encoded training/test datasets
                            - rep2
                        models: # list of reports to execute on trained ML methods for each assessment CV split
                            - rep3
                selection: # inner loop of nested CV
                    split_strategy: k_fold # perform k-fold CV
                    split_count: 5 # how many fold to create: here these two parameters mean: do 5-fold CV
                    reports:
                        data_splits: # list of reports to execute on training/test datasets (in the inner loop, so these are actually training and validation datasets)
                            - rep1
                        models: # list of reports to execute on trained ML methods for each selection CV split
                            - rep2
                        encoding: # list of reports to execute on encoded training/test datasets (again, it is training/validation here)
                            - rep3
                labels: # list of labels to optimize the classifier for, as given in the metadata for the dataset
                    - celiac:
                        positive_class: + # if it's binary classification, positive class parameter should be set
                    - T1D # this is not binary label, so no need to specify positive class
                dataset: d1 # which dataset to use for the nested CV
                strategy: GridSearch # how to choose the combinations which to test from settings (GridSearch means test all)
                metrics: # list of metrics to compute for all settings, but these do not influence the choice of optimal model
                    - accuracy
                    - auc
                reports: # list of reports to execute when nested CV is finished to show overall performance
                    - rep4
                number_of_processes: 4 # number of parallel processes to create (could speed up the computation)
                optimization_metric: balanced_accuracy # the metric to use for choosing the optimal model and during training
                refit_optimal_model: False # use trained model, do not refit on the full dataset
                export_all_ml_settings: False # only export the optimal setting
                region_type: IMGT_CDR3
                sequence_type: AMINO_ACID

    SPLITCONFIG

    REPORTCONFIG

    """

    def __init__(self, dataset, hp_strategy: HPOptimizationStrategy, hp_settings: list, assessment: SplitConfig,
                 selection: SplitConfig, metrics: set, optimization_metric: ClassificationMetric,
                 label_configuration: LabelConfiguration, path: Path = None, context: dict = None,
                 number_of_processes: int = 1, reports: dict = None, name: str = None,
                 refit_optimal_model: bool = False, sequence_type: SequenceType = None, region_type: RegionType = None,
                 export_all_ml_settings: bool = False, example_weighting: ExampleWeightingStrategy = None):
        self.state = TrainMLModelState(dataset, hp_strategy, hp_settings, assessment, selection, metrics,
                                       optimization_metric, label_configuration, path, context, number_of_processes,
                                       reports if reports is not None else {}, name, refit_optimal_model,
                                       export_all_ml_settings, example_weighting, sequence_type=sequence_type,
                                       region_type=region_type)

    def run(self, result_path: Path):
        self.state.path = result_path
        self.state = HPAssessment.run_assessment(self.state)
        self._export_all_ml_settings()
        self._compute_optimal_hp_item_per_label()
        self.state.report_results = HPUtil.run_hyperparameter_reports(self.state, self.state.path / "reports")
        self.print_performances(self.state)
        self._export_all_performances_to_csv()
        return self.state

    def _export_all_ml_settings(self):
        if self.state.export_all_ml_settings:
            for label in self.state.label_configuration.get_label_objects():
                for assessment_state in self.state.assessment_states:
                    for name, hp_item in assessment_state.label_states[label.name].assessment_items.items():
                        zip_path = MLExporter.export_zip(hp_item=hp_item, path=hp_item.ml_settings_export_path, label_name=label.name)
                        logging.info(f"TrainMLModelInstruction: config for {name} for label {label.name} was exported to: {zip_path}")

    def _compute_optimal_hp_item_per_label(self):
        n_labels = self.state.label_configuration.get_label_count()

        for idx, label in enumerate(self.state.label_configuration.get_label_objects()):
            self._compute_optimal_item(label, f"(label {idx + 1} / {n_labels})")
            zip_path = MLExporter.export_zip(hp_item=self.state.optimal_hp_items[label.name], path=self.state.path / f"optimal_{label.name}", label_name=label.name)
            self.state.optimal_hp_item_paths[label.name] = zip_path

    def _compute_optimal_item(self, label: Label, index_repr: str):
        optimal_hp_settings = [state.label_states[label.name].optimal_hp_setting for state in self.state.assessment_states]
        optimal_hp_setting = Counter(optimal_hp_settings).most_common(1)[0][0]
        if self.state.refit_optimal_model:
            print_log(f"TrainMLModel: retraining optimal model for label {label.name} {index_repr}.\n", include_datetime=True)
            self.state.optimal_hp_items[label.name] = MLProcess(self.state.dataset, None, label, self.state.metrics, self.state.optimization_metric,
                                                                self.state.path / f"optimal_{label.name}", number_of_processes=self.state.number_of_processes,
                                                                label_config=self.state.label_configuration, hp_setting=optimal_hp_setting,
                                                                example_weighting=self.state.example_weighting).run(0)
            print_log(f"TrainMLModel: finished retraining optimal model for label {label.name} {index_repr}.\n", include_datetime=True)

        else:
            optimal_assessment_state = self.state.assessment_states[optimal_hp_settings.index(optimal_hp_setting)]
            self.state.optimal_hp_items[label.name] = optimal_assessment_state.label_states[label.name].optimal_assessment_item

    def print_performances(self, state: TrainMLModelState):
        print_log(f"Performances ({state.optimization_metric.name.lower()}) -----------------------------------------------")

        for label_name in state.label_configuration.get_labels_by_name():
            print_log(f"\n\nLabel: {label_name}")
            print_log(f"Performance ({state.optimization_metric.name.lower()}) per assessment split:")
            for split in range(state.assessment.split_count):
                print_log(f"Split {split+1}: {state.assessment_states[split].label_states[label_name].optimal_assessment_item.performance[state.optimization_metric.name.lower()]}")
            if all(isinstance(a_state.label_states[label_name].optimal_assessment_item.performance[state.optimization_metric.name.lower()], float)
                   for a_state in state.assessment_states):
                print_log(f"Average performance ({state.optimization_metric.name.lower()}): "
                      f"{sum([state.assessment_states[split].label_states[label_name].optimal_assessment_item.performance[state.optimization_metric.name.lower()] for split in range(state.assessment.split_count)])/state.assessment.split_count}")
            print_log("------------------------------")

    def _export_all_performances_to_csv(self):
        self._export_optimal_performances_to_csv()
        self._export_assessment_performances_to_csv()
        if not (self.state.selection.training_percentage == 1 and self.state.selection.split_strategy == SplitType.RANDOM):
            self._export_selection_performance_to_csv()

    def _export_optimal_performances_to_csv(self):
        for label_name in self.state.label_configuration.get_labels_by_name():
            performance = {"hp_setting": [], "split": [], **{metric.name.lower(): [] for metric in self.state.metrics}}
            for index, assessment_state in enumerate(self.state.assessment_states):
                performance['split'].append(index+1)
                performance['hp_setting'].append(assessment_state.label_states[label_name].optimal_hp_setting.get_key())
                for metric in self.state.metrics:
                    performance[metric.name.lower()].append(assessment_state.label_states[label_name].optimal_assessment_item.performance[metric.name.lower()])
            pd.DataFrame(performance).to_csv(self.state.path / f"{label_name}_optimal_models_performance.csv", index=False)

    def _export_assessment_performances_to_csv(self):
        for label_name in self.state.label_configuration.get_labels_by_name():
            performance = {'hp_setting': [], 'split': [], 'optimal': [], **{metric.name.lower(): [] for metric in self.state.metrics}}
            for index, assessment_state in enumerate(self.state.assessment_states):
                for hp_setting, hp_item in assessment_state.label_states[label_name].assessment_items.items():
                    performance['hp_setting'].append(str(hp_setting))
                    performance['split'].append(index+1)
                    performance['optimal'].append(hp_setting == assessment_state.label_states[label_name].optimal_hp_setting)
                    for metric in self.state.metrics:
                        performance[metric.name.lower()].append(hp_item.performance[metric.name.lower()])
            pd.DataFrame(performance).to_csv(self.state.path / f"{label_name}_all_assessment_performances.csv", index=False)

    def _export_selection_performance_to_csv(self):
        for label_name in self.state.label_configuration.get_labels_by_name():
            for index, assessment_state in enumerate(self.state.assessment_states):
                selection_state = assessment_state.label_states[label_name].selection_state
                performance = {'hp_setting': [], 'split': [], **{metric.name.lower(): [] for metric in self.state.metrics}}
                for hp_setting, hp_item_list in selection_state.hp_items.items():
                    for i, hp_item in enumerate(hp_item_list):
                        performance['hp_setting'].append(hp_setting)
                        performance['split'].append(i+1)
                        for metric in self.state.metrics:
                            performance[metric.name.lower()].append(hp_item.performance[metric.name.lower()])
                pd.DataFrame(performance).to_csv(self.state.path / f"{label_name}_assessment_split_{index+1}_selection_performance.csv", index=False)

    @staticmethod
    def get_documentation():
        doc = str(TrainMLModelInstruction.__doc__)
        valid_values = str([metric.name.lower() for metric in ClassificationMetric])[1:-1].replace("'", "`")
        valid_strategies = str(ReflectionHandler.all_nonabstract_subclass_basic_names(HPOptimizationStrategy, "",
                                                                                      "hyperparameter_optimization/strategy/"))[1:-1]\
            .replace("'", "`")
        mapping = {
            "dataset (Dataset)": "dataset",
            "hp_strategy (HPOptimizationStrategy)": "strategy",
            "hp_settings": "settings",
            "assessment (SplitConfig)": "assessment",
            "selection (SplitConfig)": "selection",
            "optimization_metric (Metric)": "optimization_metric",
            "label_configuration (LabelConfiguration)": "labels (list)",
            "data_reports": "reports",
            "a list of metrics": f"a list of metrics ({valid_values})",
            "a metric to use for optimization": f"a metric to use for optimization (one of {valid_values})",
            "Valid values are objects of any class inheriting :py:obj:`~immuneML.hyperparameter_optimization.strategy."
            "HPOptimizationStrategy.HPOptimizationStrategy`.": f"Valid values are: {valid_strategies}.",
            "the reports to be specified here have to be :py:obj:`~immuneML.reports.train_ml_model_reports.TrainMLModelReport.TrainMLModelReport` reports.": f"the reports that can be provided here are :ref:`**{TrainMLModelReport.DOCS_TITLE}**`.",
            "SPLITCONFIG": SplitConfig.get_documentation(),
            "REPORTCONFIG": ReportConfig.get_documentation()
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import dataclasses
from typing import Union

from olga.load_model import GenerativeModelVJ, GenerativeModelVDJ, GenomicData
from olga.sequence_generation import SequenceGenerationVDJ, SequenceGenerationVJ


@dataclasses.dataclass
class InternalOlgaModel:
    sequence_gen_model: Union[SequenceGenerationVDJ, SequenceGenerationVJ] = None
    v_gene_mapping: list = None
    j_gene_mapping: list = None
    genomic_data: GenomicData = None
    olga_gen_model: Union[GenerativeModelVDJ, GenerativeModelVJ] = None



import abc
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Union

from immuneML.data_model.EncodedData import EncodedData


@dataclass
class Dataset:
    identifier: str = None
    name: str = None
    encoded_data: EncodedData = None
    labels: dict = field(default_factory=dict)
    dataset_file: Path = None

    TRAIN = "train"
    TEST = "test"
    SUBSAMPLED = "subsampled"
    PREPROCESSED = "preprocessed"

    def __post_init__(self):
        if self.name is None:
            self.name = self.identifier

    @classmethod
    @abc.abstractmethod
    def build_from_objects(cls, **kwargs):
        pass

    @classmethod
    @abc.abstractmethod
    def create_metadata_dict(cls, **kwargs):
        pass

    @abc.abstractmethod
    def make_subset(self, example_indices, path, dataset_type: str):
        pass

    @abc.abstractmethod
    def get_example_count(self):
        pass

    @abc.abstractmethod
    def get_data(self, batch_size: int = 1):
        pass

    @abc.abstractmethod
    def get_locus(self):
        pass

    @abc.abstractmethod
    def get_example_ids(self):
        pass

    @abc.abstractmethod
    def get_label_names(self):
        pass

    @abc.abstractmethod
    def clone(self, keep_identifier: bool = False):
        pass

    @abc.abstractmethod
    def get_metadata(self, field_names: Union[list, None], return_df: bool = False):
        pass

    @abc.abstractmethod
    def get_data_from_index_range(self, start_index: int, end_index: int):
        pass


from dataclasses import fields
from typing import Dict

import bionumpy as bnp
from bionumpy import AminoAcidEncoding
from bionumpy.bnpdataclass import bnpdataclass
from bionumpy.encodings import AlphabetEncoding
from numpy._typing import ArrayLike

AminoAcidXEncoding = AlphabetEncoding('ACDEFGHIKLMNPQRSTVWXY*')
DNANEncoding = AlphabetEncoding("ACGTN")


@bnpdataclass
class AIRRSequenceSet:
    sequence_id: str = ''
    sequence: DNANEncoding = ''
    quality: str = ''
    sequence_aa: AminoAcidXEncoding = ''
    rev_comp: str = ''
    productive: str = ''
    vj_in_frame: str = ''
    stop_codon: str = ''
    complete_vdj: str = ''
    locus: str = ''
    locus_species: str = ''
    v_call: str = ''
    d_call: str = ''
    d2_call: str = ''
    j_call: str = ''
    c_call: str = ''
    sequence_alignment: str = ''
    quality_alignment: str = ''
    sequence_alignment_aa: str = ''
    germline_alignment: str = ''
    germline_alignment_aa: str = ''
    junction: DNANEncoding = None
    junction_aa: AminoAcidXEncoding = None
    np1: DNANEncoding = None
    np1_aa: AminoAcidXEncoding = None
    np2: DNANEncoding = None
    np2_aa: AminoAcidXEncoding = None
    np3: DNANEncoding = None
    np3_aa: AminoAcidXEncoding = None
    cdr1: DNANEncoding = None
    cdr1_aa: AminoAcidXEncoding = None
    cdr2: DNANEncoding = None
    cdr2_aa: AminoAcidXEncoding = None
    cdr3: DNANEncoding = None
    cdr3_aa: AminoAcidXEncoding = None
    fwr1: DNANEncoding = None
    fwr1_aa: AminoAcidXEncoding = None
    fwr2: DNANEncoding = None
    fwr2_aa: AminoAcidXEncoding = None
    fwr3: DNANEncoding = None
    fwr3_aa: AminoAcidXEncoding = None
    fwr4: DNANEncoding = None
    fwr4_aa: AminoAcidXEncoding = None
    v_score: float = None
    v_identity: float = None
    v_support: float = None
    v_cigar: str = ''
    d_score: float = None
    d_identity: float = None
    d_support: float = None
    d_cigar: str = ''
    d2_score: float = None
    d2_identity: float = None
    d2_support: float = None
    d2_cigar: str = ''
    j_score: float = None
    j_identity: float = None
    j_support: float = None
    j_cigar: str = ''
    c_score: float = None
    c_identity: float = None
    c_support: float = None
    c_cigar: str = ''
    v_sequence_start: int = None
    v_sequence_end: int = None
    v_germline_start: int = None
    v_germline_end: int = None
    v_alignment_start: int = None
    v_alignment_end: int = None
    d_sequence_start: int = None
    d_sequence_end: int = None
    d_germline_start: int = None
    d_germline_end: int = None
    d_alignment_start: int = None
    d_alignment_end: int = None
    d2_sequence_start: int = None
    d2_sequence_end: int = None
    d2_germline_start: int = None
    d2_germline_end: int = None
    d2_alignment_start: int = None
    d2_alignment_end: int = None
    j_sequence_start: int = None
    j_sequence_end: int = None
    j_germline_start: int = None
    j_germline_end: int = None
    j_alignment_start: int = None
    j_alignment_end: int = None
    c_sequence_start: int = None
    c_sequence_end: int = None
    c_germline_start: int = None
    c_germline_end: int = None
    c_alignment_start: int = None
    c_alignment_end: int = None
    cdr1_start: int = None
    cdr1_end: int = None
    cdr2_start: int = None
    cdr2_end: int = None
    cdr3_start: int = None
    cdr3_end: int = None
    fwr1_start: int = None
    fwr1_end: int = None
    fwr2_start: int = None
    fwr2_end: int = None
    fwr3_start: int = None
    fwr3_end: int = None
    fwr4_start: int = None
    fwr4_end: int = None
    v_sequence_alignment: str = ''
    v_sequence_alignment_aa: str = ''
    d_sequence_alignment: str = ''
    d_sequence_alignment_aa: str = ''
    d2_sequence_alignment: str = ''
    d2_sequence_alignment_aa: str = ''
    j_sequence_alignment: str = ''
    j_sequence_alignment_aa: str = ''
    c_sequence_alignment: str = ''
    c_sequence_alignment_aa: str = ''
    v_germline_alignment: str = ''
    v_germline_alignment_aa: str = ''
    d_germline_alignment: str = ''
    d_germline_alignment_aa: str = ''
    d2_germline_alignment: str = ''
    d2_germline_alignment_aa: str = ''
    j_germline_alignment: str = ''
    j_germline_alignment_aa: str = ''
    c_germline_alignment: str = ''
    c_germline_alignment_aa: str = ''
    junction_length: int = None
    junction_aa_length: int = None
    np1_length: int = None
    np2_length: int = None
    np3_length: int = None
    n1_length: int = None
    n2_length: int = None
    n3_length: int = None
    p3v_length: int = None
    p5d_length: int = None
    p3d_length: int = None
    p5d2_length: int = None
    p3d2_length: int = None
    p5j_length: int = None
    v_frameshift: str = ''
    j_frameshift: str = ''
    d_frame: int = None
    d2_frame: int = None
    consensus_count: int = None
    duplicate_count: int = None
    umi_count: int = None
    cell_id: str = ''
    clone_id: str = ''
    repertoire_id: str = ''
    sample_processing_id: str = ''
    data_processing_id: str = ''
    rearrangement_id: str = ''
    rearrangement_set_id: str = ''
    germline_database: str = ''

    STR_TO_TYPE = {'str': str, 'int': int, 'float': float, 'bool': bool,
                   'AminoAcidEncoding': bnp.encodings.AminoAcidEncoding,
                   'AminoAcidXEncoding': AminoAcidXEncoding,
                   'DNANEncoding': DNANEncoding}

    TYPE_TO_STR = {**{val: key for key, val in STR_TO_TYPE.items()},
                   **{AlphabetEncoding('ACDEFGHIKLMNPQRSTVWY*'): 'AminoAcidEncoding',
                      AlphabetEncoding('ACGTN'): 'DNANEncoding',
                      AlphabetEncoding('ACDEFGHIKLMNPQRSTVWXY*'): 'AminoAcidXEncoding'}}

    @classmethod
    def get_neutral_value(cls, field_type):
        neutral_values = {str: '', int: -1, DNANEncoding: '', AminoAcidEncoding: '', AminoAcidXEncoding: '', float: -1.,
                          bool: ''}
        return neutral_values[field_type]

    @classmethod
    def get_field_type_dict(cls, all_fields: bool = True):
        if all_fields:
            return {f.name: f.type for f in fields(cls)}
        else:
            return {f.name: f.type for f in fields(cls)
                    if f.name not in [airr_f.name for airr_f in fields(AIRRSequenceSet)]}

    def add_fields(self, fields: Dict[str, ArrayLike], field_type_map: dict):
        fields_with_types = [(field_name, field_type, self.__class__.get_neutral_value(field_type))
                             for field_name, field_type in field_type_map.items()]
        new_class = self.__class__.extend(fields_with_types)
        return new_class(**{**vars(self), **fields})

    @classmethod
    def get_dynamic_fields(cls):
        standard_fields = [f.name for f in fields(AIRRSequenceSet)]
        return {f.name: f.type for f in fields(cls) if f not in standard_fields}



import numpy as np
from sklearn import metrics


def brier_score_loss(true_y, predicted_y, sample_weight=None, labels=None):
    return metrics.brier_score_loss(true_y, predicted_y, sample_weight=sample_weight)


def precision_score_micro(true_y, predicted_y, sample_weight=None, labels=None):
    return metrics.precision_score(true_y, predicted_y, average="micro", sample_weight=sample_weight, labels=labels)


def precision_score_macro(true_y, predicted_y, sample_weight=None, labels=None):
    return metrics.precision_score(true_y, predicted_y, average="macro", sample_weight=sample_weight, labels=labels)


def precision_score_weighted(true_y, predicted_y, sample_weight=None, labels=None):
    return metrics.precision_score(true_y, predicted_y, average="weighted", sample_weight=sample_weight, labels=labels)


def recall_score_micro(true_y, predicted_y, sample_weight=None, labels=None):
    return metrics.recall_score(true_y, predicted_y, average="micro", sample_weight=sample_weight, labels=labels)


def recall_score_macro(true_y, predicted_y, sample_weight=None, labels=None):
    return metrics.recall_score(true_y, predicted_y, average="macro", sample_weight=sample_weight, labels=labels)


def recall_score_weighted(true_y, predicted_y, sample_weight=None, labels=None):
    return metrics.recall_score(true_y, predicted_y, average="weighted", sample_weight=sample_weight, labels=labels)


def f1_score_weighted(true_y, predicted_y, sample_weight=None):
    return metrics.f1_score(true_y, predicted_y, average="weighted", sample_weight=sample_weight)


def f1_score_micro(true_y, predicted_y, sample_weight=None):
    return metrics.f1_score(true_y, predicted_y, average="micro", sample_weight=sample_weight)


def f1_score_macro(true_y, predicted_y, sample_weight=None):
    return metrics.f1_score(true_y, predicted_y, average="macro", sample_weight=sample_weight)


def roc_auc_score_ovo(true_y, predicted_y, sample_weight=None, labels=None):
    return roc_auc_score(true_y, predicted_y, sample_weight=sample_weight, labels=labels, multiclass='ovo')


def roc_auc_score_ovr(true_y, predicted_y, sample_weight=None, labels=None):
    return roc_auc_score(true_y, predicted_y, sample_weight=sample_weight, labels=labels, multiclass='ovr')


def roc_auc_score(true_y, predicted_y, sample_weight=None, labels=None, multiclass: str = 'raise'):
    predictions = np.array(predicted_y) if not isinstance(predicted_y, np.ndarray) else predicted_y
    true_values = np.array(true_y) if not isinstance(true_y, np.ndarray) else true_y
    if predictions.shape == true_values.shape:
        return metrics.roc_auc_score(true_values, predictions, sample_weight=sample_weight, labels=labels,
                                     multi_class=multiclass)
    elif len(predictions.shape) == 2 and predictions.shape[1] == 2:
        return metrics.roc_auc_score(true_values, predictions[:, 1], sample_weight=sample_weight, labels=labels,
                                     multi_class=multiclass)
    else:
        return -1



from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pandas as pd

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod


class DataFrameWrapper:

    def __init__(self, path: Path, df: pd.DataFrame = None):
        self.path = path
        self.df = df

        if df is not None and not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            df.to_csv(str(path), index=False)

    def get_df(self):
        if self.df is None and isinstance(self.path, Path) and self.path.exists():
            self.df = pd.read_csv(str(self.path))
        return self.df


@dataclass
class ClusteringSetting:
    encoder: DatasetEncoder
    encoder_params: dict
    encoder_name: str
    clustering_method: ClusteringMethod
    clustering_params: dict
    clustering_method_name: str
    dim_reduction_method: DimRedMethod = None
    dim_red_params: dict = None
    dim_red_name: str = None
    path: Path = None

    def get_key(self) -> str:
        key = self.encoder_name
        if self.dim_red_name:
            key += f"_{self.dim_red_name}"
        key += f"_{self.clustering_method_name}"
        return key

    def __str__(self):
        return self.get_key()

    def __hash__(self):
        return hash(self.get_key())

    def __eq__(self, other):
        return self.get_key() == other.get_key()


@dataclass
class ClusteringItem:
    dataset: Dataset = None
    method: ClusteringMethod = None
    encoder: DatasetEncoder = None
    internal_performance: DataFrameWrapper = None
    external_performance: DataFrameWrapper = None
    predictions: np.ndarray = None
    cl_setting: ClusteringSetting = None



from random import randrange

import numpy as np
import pandas as pd

from immuneML.analysis.criteria_matches.BooleanType import BooleanType
from immuneML.analysis.criteria_matches.OperationType import OperationType


class CriteriaMatcher:
    """
    Takes a data frame (for example, repertoire or feature annotations) and criteria and allowed values
    as input and returns a list of boolean values indicating a match or not for each row

    """

    def match(self, criteria, data):
        return CriteriaMatcher.parse_criteria(criteria, data)

    @staticmethod
    def evaluate_in(data: pd.Series, criteria: dict):
        result = data.isin(criteria["values"])
        return result.values

    @staticmethod
    def evaluate_not_in(data: pd.Series, criteria: dict):
        result = ~data.isin(criteria["values"])
        return result.values

    @staticmethod
    def evaluate_not_na(data: pd.Series, criteria: dict):
        result = data.notna()
        return result.values

    @staticmethod
    def evaluate_greater_than(data: pd.Series, criteria: dict):
        result = data > criteria["threshold"]
        return result.values

    @staticmethod
    def evaluate_less_than(data: pd.Series, criteria: dict):
        result = data < criteria["threshold"]
        return result.values

    @staticmethod
    def evaluate_top_n(data: pd.Series, criteria: dict):
        top_n = data.values.argsort()[(-1 * criteria["number"]):][::-1]
        result = [i in top_n for i in range(data.size)]
        return np.array(result)

    @staticmethod
    def evaluate_random_n(data: pd.Series, criteria: dict):
        random_n = [randrange(0, data.size) for i in range(criteria["number"])]
        result = [i in random_n for i in range(data.size)]
        return np.array(result)

    @staticmethod
    def evaluate_and(booleans: list):
        result = np.logical_and.reduce(booleans)
        return result

    @staticmethod
    def evaluate_or(booleans: list):
        result = np.logical_or.reduce(booleans)
        return result

    @staticmethod
    def evaluate_column(data: pd.Series, name: str) -> pd.Series:
        return data[name]

    @staticmethod
    def parse_criteria(criteria, data):
        if criteria["type"] in OperationType:
            operation = getattr(CriteriaMatcher, "evaluate_" + criteria["type"].name.lower())
            return operation(data[criteria['column']], criteria)
        elif criteria["type"] in BooleanType:
            operation = getattr(CriteriaMatcher, "evaluate_" + criteria["type"].name.lower())
            booleans = []
            for operand in criteria["operands"]:
                booleans.append(CriteriaMatcher.parse_criteria(operand, data))
            return operation(booleans)



import copy
import random
import shutil
from pathlib import Path
from typing import List, Union

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.IO.dataset_export.DataExporter import DataExporter
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.bnp_util import bnp_write_to_file, write_yaml, read_yaml
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.subsampling.SubsamplingState import SubsamplingState
from scripts.specification_util import update_docs_per_mapping


class SubsamplingInstruction(Instruction):
    """
    Subsampling is an instruction that subsamples a given dataset and creates multiple smaller dataset according to the
    parameters provided.

    **Specification arguments:**

    - dataset (str): original dataset which will be used as a basis for subsampling

    - subsampled_dataset_sizes (list): a list of dataset sizes (number of examples) each subsampled dataset should have

    - subsampled_repertoire_size (int): the number of sequences to keep per repertoire (or None if all sequences should
      be kept) if dataset is a RepertoireDataset; otherwise, this argument is ignored.

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_subsampling_instruction: # user-defined name of the instruction
                type: Subsampling # which instruction to execute
                dataset: my_dataset # original dataset to be subsampled, with e.g., 300 examples
                subsampled_dataset_sizes: # how large the subsampled datasets should be, one dataset will be created for each list item
                    - 200 # one subsampled dataset with 200 examples (200 repertoires if my_dataset was repertoire dataset)
                    - 100 # the other subsampled dataset will have 100 examples

    """

    def __init__(self, dataset: Dataset, subsampled_dataset_sizes: List[int],
                 subsampled_repertoire_size: Union[int, None] = None, result_path: Path = None, name: str = None):
        self.state = SubsamplingState(dataset, subsampled_dataset_sizes, subsampled_repertoire_size, result_path, name)

    def run(self, result_path: Path):
        self.state.result_path = PathBuilder.build(result_path / self.state.name)

        example_indices = list(range(self.state.dataset.get_example_count()))

        for index, dataset_size in enumerate(self.state.subsampled_dataset_sizes):

            new_dataset_name = f"{self.state.dataset.name}_{dataset_size}_subsampled_{index+1}"
            new_dataset_path = PathBuilder.build(self.state.result_path / new_dataset_name)

            new_example_indices = random.sample(example_indices, k=dataset_size)

            if isinstance(self.state.dataset, RepertoireDataset) and self.state.subsampled_repertoire_size is not None:
                new_dataset = self.make_repertoire_dataset_subset(new_example_indices, new_dataset_path, new_dataset_name)
            else:
                new_dataset = self.state.dataset.make_subset(new_example_indices, new_dataset_path, Dataset.SUBSAMPLED)
                new_dataset.name = new_dataset_name

            self.state.subsampled_datasets.append(new_dataset)

            self.export_dataset(new_dataset, new_dataset_path)

        return self.state

    def make_repertoire_dataset_subset(self, new_example_indices: list, new_dataset_path: Path, name: str):
        repertoires = []

        for index in new_example_indices:
            repertoire = self.state.dataset.repertoires[index]
            metadata = copy.deepcopy(repertoire.metadata) if repertoire.metadata is not None \
                else read_yaml(repertoire.metadata_filename)

            data = repertoire.data

            if self.state.subsampled_repertoire_size:
                rows_to_keep = random.sample(list(range(len(data))), k=self.state.subsampled_repertoire_size)
                data = data[rows_to_keep]

            rep_name = f"{repertoire.data_filename.stem}_subsampled_{self.state.subsampled_repertoire_size}"
            data_filename = new_dataset_path / f"{rep_name}.tsv"
            metadata_filename = new_dataset_path / f"{rep_name}.yaml"

            bnp_write_to_file(data_filename, data)
            write_yaml(metadata_filename, metadata)

            repertoires.append(Repertoire(data_filename, metadata_filename))

        return RepertoireDataset.build_from_objects(repertoires=repertoires, path=new_dataset_path, name=name)

    def export_dataset(self, new_dataset, new_dataset_path):
        self.state.subsampled_dataset_paths[new_dataset.name] = {}
        exporter = AIRRExporter
        exporter_name = exporter.__name__[:-8].lower()
        export_path = new_dataset_path / f"exported/{exporter_name}/"
        exporter.export(new_dataset, export_path)
        zip_export_path = shutil.make_archive(new_dataset_path / f"exported_{exporter_name}_{new_dataset.name}", "zip", export_path)
        self.state.subsampled_dataset_paths[new_dataset.name][exporter_name] = zip_export_path

    @staticmethod
    def get_documentation():
        doc = str(SubsamplingInstruction.__doc__)

        valid_strategy_values = ReflectionHandler.all_nonabstract_subclass_basic_names(DataExporter, "Exporter", "dataset_export/")
        valid_strategy_values = str(valid_strategy_values)[1:-1].replace("'", "`")
        mapping = {
            "Valid formats are class names of any non-abstract class inheriting "
            ":py:obj:`~immuneML.IO.dataset_export.DataExporter.DataExporter`.": f"Valid values are: {valid_strategy_values}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import logging

from immuneML import Constants
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingStrategy import SequenceEncodingStrategy
from immuneML.util.KmerHelper import KmerHelper


class VGeneIMGTKmerEncoder(SequenceEncodingStrategy):

    @staticmethod
    def encode_sequence(sequence: ReceptorSequence, params: EncoderParams, encode_locus=False):
        """
        Encodes a receptor sequence into a sequence of k-mers

        Args:
            encode_locus:
            sequence: ReceptorSequence object
            params: EncoderParams object with information on k-mer length

        Returns:

        """
        k = params.model["k"]
        sequence_type = params.model.get('sequence_type', params.sequence_type)

        length = len(sequence.get_sequence(sequence_type))

        if length < k:
            logging.warning(f'KmerSequenceEncoder: Sequence length {length} is less than {k}. Ignoring sequence...')
            return None

        kmers = KmerHelper.create_IMGT_kmers_from_sequence(sequence=sequence, k=k, sequence_type=sequence_type,
                                                           region_type=params.model.get('region_type',
                                                                                        params.region_type))

        kmers = [Constants.FEATURE_DELIMITER.join([sequence.v_gene] + [str(mer) for mer in kmer]) for kmer in kmers]
        if encode_locus:
            kmers = [f"{sequence.locus}{Constants.FEATURE_DELIMITER}{kmer}" for kmer in kmers]

        return kmers

    @staticmethod
    def get_feature_names(params: EncoderParams):
        return ["sequence"]



from immuneML.dsl.ObjectParser import ObjectParser
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.reports.Report import Report
from immuneML.util.Logger import log
from immuneML.util.ReflectionHandler import ReflectionHandler


class ReportParser:

    keyword = "reports"

    @staticmethod
    def parse(reports: dict, symbol_table: SymbolTable):
        if reports is None or len(reports) == 0:
            reports = {}

        for rep_id in reports.keys():
            symbol_table, reports[rep_id] = ReportParser._parse_report(rep_id, reports[rep_id], symbol_table)

        return symbol_table, reports

    @staticmethod
    def _parse_report(key: str, params: dict, symbol_table: SymbolTable):
        valid_values = ReflectionHandler.all_nonabstract_subclass_basic_names(Report, "", "reports/")
        report_object, params = ObjectParser.parse_object(params, valid_values, "", "reports/", "ReportParser", key, builder=True,
                                                          return_params_dict=True)

        symbol_table.add(key, SymbolType.REPORT, report_object)

        return symbol_table, params



import abc
import logging
from abc import ABC
from typing import List

import numpy as np

from immuneML.data_model.datasets.Dataset import Dataset


class DimRedMethod(ABC):
    """
    Dimensionality reduction methods are algorithms which can be used to reduce the dimensionality
    of encoded datasets, in order to uncover and analyze patterns present in the data.

    These methods can be used in the :ref:`ExploratoryAnalysis` and :ref:`Clustering` instructions.
    """

    DOCS_TITLE = "Dimensionality reduction methods"

    def __init__(self, name: str = None):
        self.method = None
        self.name = name

    def fit(self, dataset: Dataset = None, design_matrix: np.ndarray = None):
        if dataset is None:
            self.method.fit(design_matrix)
        else:
            self.method.fit(dataset.encoded_data.get_examples_as_np_matrix())

    def transform(self, dataset: Dataset = None, design_matrix: np.ndarray = None):
        if dataset is None:
            return self.method.transform(design_matrix)
        else:
            return self.method.transform(dataset.encoded_data.get_examples_as_np_matrix())

    def fit_transform(self, dataset: Dataset = None, design_matrix: np.ndarray = None):
        if dataset is None:
            return self.method.fit_transform(design_matrix)
        else:
            return self.method.fit_transform(dataset.encoded_data.get_examples_as_np_matrix())

    def inverse_transform(self, transformed_data):
        try:
            return self.method.inverse_transform(transformed_data)
        except Exception:
            logging.warning(f"{self.__class__.__name__}: inverse transformation is not supported by this method.")
            return None

    @abc.abstractmethod
    def get_dimension_names(self) -> List[str]:
        pass

    @classmethod
    def get_title(cls):
        return "Dimensionality Reduction"



from pathlib import Path

import numpy as np
import plotly.express as px
import plotly.figure_factory as ff
import plotly.graph_objects as go
from scipy.cluster.hierarchy import cophenet
from scipy.spatial.distance import squareform

from immuneML.ml_methods.clustering.AgglomerativeClustering import AgglomerativeClustering
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.clustering_method_reports.ClusteringMethodReport import ClusteringMethodReport
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringItem


class Dendrogram(ClusteringMethodReport):
    """
    This report generates a dendrogram visualization from the AgglomerativeClustering method and shows the external
    labels as annotations.

    **Specification arguments:**

    - labels (list): List of metadata labels to annotate on the dendrogram.

    **YAML specification:**

    .. code-block:: yaml

        reports:
            my_dendrogram_report:
                Dendrogram:
                    labels:
                        - disease_status
                        - age_group

    """

    def __init__(self, labels: list, result_path: Path = None, name: str = None,
                 clustering_item: ClusteringItem = None):
        super().__init__(name=name, result_path=result_path)
        self.item = clustering_item
        self.labels = labels

    @classmethod
    def build_object(cls, **kwargs):
        return cls(**kwargs)

    def check_prerequisites(self) -> bool:
        return (isinstance(self.item.method, AgglomerativeClustering) and
                self.item.method.model.distance_threshold == 0 and self.item.method.model.n_clusters is None)

    def _get_linkage_matrix(self):
        counts = np.zeros(self.item.method.model.children_.shape[0])
        n_samples = len(self.item.method.model.labels_)
        for i, merge in enumerate(self.item.method.model.children_):
            current_count = 0
            for child_idx in merge:
                if child_idx < n_samples:
                    current_count += 1  # leaf node
                else:
                    current_count += counts[child_idx - n_samples]
            counts[i] = current_count

        linkage_matrix = np.column_stack(
            [self.item.method.model.children_, self.item.method.model.distances_, counts]
        ).astype(float)

        return linkage_matrix

    def _generate(self) -> ReportResult:

        # Generate output path
        output_path = Path(self.result_path) / "dendrogram.html"

        self._create_full_dendrogram(output_path)

        return ReportResult(output_figures=[ReportOutput(output_path, "Dendrogram visualization")])

    def _save_data(self, linkage_matrix, metadata):
        np.save(f"{self.result_path}/linkage_matrix.npy", linkage_matrix)
        metadata.reset_index().to_csv(f"{self.result_path}/metadata.csv", index=False)
        np.save(f"{self.result_path}/distances.npy", self.item.method.model.distances_)

    def _add_annotations(self, fig, annotation_start, annotation_height, metadata, ordered_ids):
        for i, label in enumerate(self.labels):
            yaxis_name = f'y{i + 3}'
            y_domain = [
                annotation_start + i * annotation_height,
                annotation_start + i * annotation_height + annotation_height
            ]

            annotation_values = [metadata.loc[ind, label] for ind in ordered_ids]
            unique_labels = sorted(metadata[label].unique().tolist())
            label_to_num = {label_val: j for j, label_val in enumerate(unique_labels)}
            annotation_numeric = [label_to_num[v] for v in annotation_values]

            # Add annotation heatmap
            fig.add_trace(go.Heatmap(
                x=fig['layout']['xaxis']['tickvals'],
                y=[label],
                z=np.array(annotation_numeric).reshape(1, -1),
                showlegend=False,
                showscale=False,
                colorscale=px.colors.qualitative.Plotly[1:],
                hovertemplate=label + ': %{text}<br>Example: %{x}<extra></extra>',
                text=[annotation_values],
                yaxis=yaxis_name
            ))

            # Update layout for this axis
            fig.update_layout(**{
                f'yaxis{i + 3}': dict(
                    domain=y_domain,
                    mirror=False,
                    showgrid=False,
                    showline=False,
                    zeroline=False
                )
            })

        return fig

    def _update_layout(self, fig, example_ids):

        fig.update_layout({
            'height': 700 + len(example_ids) / 10 + len(self.labels) * 20,
            # Adjust height based on number of labels and number of examples
            'autosize': True,
            'hovermode': 'closest',
            'template': 'plotly_white',
            'xaxis': {
                'domain': [0.15, 1],
                'mirror': False,
                'showgrid': False,
                'showline': False,
                'zeroline': False,
                'ticks': "",
                'showticklabels': False
            },
            'xaxis2': {
                'domain': [0, 0.145],  # Side dendrogram
                'mirror': False,
                'showgrid': False,
                'showline': False,
                'zeroline': False,
                'ticks': "",
                'showticklabels': False
            },
            'yaxis': {
                'domain': [0, 0.65],  # Reduced height for main heatmap
                'mirror': False,
                'showgrid': False,
                'showline': False,
                'zeroline': False,
                'showticklabels': False,
                'ticks': ""
            },
            'yaxis2': {
                'domain': [0.8, 1],  # Top dendrogram
                'mirror': False,
                'showgrid': False,
                'showline': False,
                'zeroline': False,
                'showticklabels': False,
                'ticks': ""
            }
        })

        return fig

    def _create_full_dendrogram(self, output_path):
        linkage_matrix = self._get_linkage_matrix()
        example_ids = self.item.dataset.get_example_ids()
        metadata = self.item.dataset.get_metadata(self.labels, return_df=True)
        metadata['example_id'] = example_ids
        metadata = metadata.set_index('example_id')

        self._save_data(linkage_matrix, metadata)

        fig, dendro_leaves, dendro_side = self._make_dendrograms(example_ids, linkage_matrix)
        fig, ordered_ids = self._add_distance_heatmap(example_ids, linkage_matrix, dendro_leaves, dendro_side, fig)

        annotation_start, annotation_end = 0.65, 0.8
        annotation_height = (annotation_end - annotation_start) / len(self.labels)

        fig = self._add_annotations(fig, annotation_start, annotation_height, metadata, ordered_ids)
        fig = self._update_layout(fig, example_ids)

        fig_path = PlotlyUtil.write_image_to_file(fig, output_path, len(example_ids))

        return fig_path

    def _add_distance_heatmap(self, example_ids, linkage_matrix, dendro_leaves, dendro_side, fig):
        ordered_ids = [example_ids[idx] for idx in dendro_leaves]
        heat_data = squareform(cophenet(linkage_matrix))
        heat_data = heat_data[dendro_leaves, :]
        heat_data = heat_data[:, dendro_leaves]

        heatmap = go.Heatmap(x=dendro_leaves, y=dendro_leaves, z=heat_data, colorscale=px.colors.sequential.Blues,
                             showlegend=False, showscale=False,
                             hovertemplate='Distance: %{z}<extra></extra>')

        heatmap['x'] = fig['layout']['xaxis']['tickvals']
        heatmap['y'] = dendro_side['layout']['yaxis']['tickvals']

        fig.add_trace(heatmap)

        return fig, ordered_ids


    def _make_dendrograms(self, example_ids, linkage_matrix):

        fig = ff.create_dendrogram(np.zeros(len(example_ids)), orientation='bottom',
                                   colorscale=['#7393B3' for _ in range(8)],
                                   labels=example_ids, linkagefun=lambda x: linkage_matrix,
                                   distfun=lambda x: linkage_matrix[:, 2])

        for i in range(len(fig['data'])):
            fig['data'][i]['yaxis'] = 'y2'
            fig['data'][i]['hovertemplate'] = None
            fig['data'][i]['showlegend'] = False

        dendro_side = ff.create_dendrogram(np.zeros(len(example_ids)), orientation='right',
                                           linkagefun=lambda x: linkage_matrix, distfun=lambda x: linkage_matrix[:, 2],
                                           colorscale=['#7393B3' for _ in range(8)])
        for i in range(len(dendro_side['data'])):
            dendro_side['data'][i]['xaxis'] = 'x2'
            dendro_side['data'][i]['hovertemplate'] = None
            dendro_side['data'][i]['showlegend'] = False

        for data in dendro_side['data']:
            fig.add_trace(data)

        dendro_leaves = dendro_side['layout']['yaxis']['ticktext']
        dendro_leaves = list(map(int, dendro_leaves))

        return fig, dendro_leaves, dendro_side



from enum import Enum


class SequenceEncodingType(Enum):
    GAPPED_KMER = "GappedKmerSequenceEncoder"
    IMGT_CONTINUOUS_KMER = "IMGTKmerSequenceEncoder"
    CONTINUOUS_KMER = "KmerSequenceEncoder"
    IMGT_GAPPED_KMER = "IMGTGappedKmerEncoder"
    V_GENE_CONT_KMER = "VGeneContKmerEncoder"
    V_GENE_IMGT_KMER = 'VGeneIMGTKmerEncoder'



import logging

from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType


class DefinitionParserOutput:

    def __init__(self, symbol_table: SymbolTable, specification: dict):
        if not any(len(symbol_table.get_by_type(symbol_type)) > 0 for symbol_type in [t for t in SymbolType]):
            logging.warning("DefinitionParserOutput: symbol table has not been populated by objects of any type."
                          "Please check if 'definitions' are correctly specified. ")

        self.symbol_table = symbol_table
        self.specification = specification



import logging
from numbers import Number
from pathlib import Path

import pandas as pd
import plotly.express as px
import yaml

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.GradientBoosting import GradientBoosting
from immuneML.ml_methods.classifiers.LogisticRegression import LogisticRegression
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.classifiers.RandomForestClassifier import RandomForestClassifier
from immuneML.ml_methods.classifiers.SVC import SVC
from immuneML.ml_methods.classifiers.SVM import SVM
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.CoefficientPlottingSetting import CoefficientPlottingSetting
from immuneML.reports.ml_reports.CoefficientPlottingSettingList import CoefficientPlottingSettingList
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from scripts.specification_util import update_docs_per_mapping


class Coefficients(MLReport):
    """
    A report that plots the coefficients for a given ML method in a barplot. Can be used for :ref:`LogisticRegression`,
    :ref:`SVM`, :ref:`SVC`, and :ref:`RandomForestClassifier`. In the case of RandomForest, the feature importances will be plotted.

    When used in :ref:`TrainMLModel` instruction, the report can be specified under 'models', both on
    the selection and assessment levels.

    Which coefficients should be plotted (for example: only nonzero, above a certain threshold, ...) can be specified.
    Multiple options can be specified simultaneously. By default the 25 largest coefficients are plotted.
    The full set of coefficients will also be exported as a csv file.


    Example output:

    .. image:: ../../_static/images/reports/coefficients_logistic_regression.png
       :alt: Coefficients report
       :width: 650


    **Specification arguments:**

    - coefs_to_plot (list): A list specifying which coefficients should be plotted. For options see :py:obj:`~immuneML.reports.ml_reports.CoefficientPlottingSetting.CoefficientPlottingSetting`.

    - cutoff (list): If 'cutoff' is specified under 'coefs_to_plot', the cutoff values can be specified here. The coefficients which have an absolute value equal to or greater than the cutoff will be plotted.

    - n_largest (list): If 'n_largest' is specified under 'coefs_to_plot', the values for n can be specified here. These should be integer values. The n largest coefficients are determined based on their absolute values.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_coef_report:
                    Coefficients:
                        coefs_to_plot:
                            - all
                            - nonzero
                            - cutoff
                            - n_largest
                        cutoff:
                            - 0.1
                            - 0.01
                        n_largest:
                            - 5
                            - 10

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = "Coefficients"
        coefs_to_plot = [coef.upper() for coef in kwargs["coefs_to_plot"]]

        name = kwargs["name"] if "name" in kwargs else None

        ParameterValidator.assert_all_in_valid_list(coefs_to_plot, [item.name.upper() for item in CoefficientPlottingSetting], location,
                                                    "coefs_to_plot")

        if CoefficientPlottingSetting.CUTOFF.name in coefs_to_plot:
            cutoff = kwargs["cutoff"]
            ParameterValidator.assert_type_and_value(cutoff, list, location, "cutoff")
            ParameterValidator.assert_all_type_and_value(cutoff, Number, location, "cutoff", min_inclusive=1e-15)
        else:
            cutoff = []

        if CoefficientPlottingSetting.N_LARGEST.name in coefs_to_plot:
            n_largest = kwargs["n_largest"]
            ParameterValidator.assert_type_and_value(n_largest, list, location, "n_largest")
            ParameterValidator.assert_all_type_and_value(n_largest, int, location, "n_largest", min_inclusive=1)
        else:
            n_largest = []

        coefs = CoefficientPlottingSettingList()
        for keyword in coefs_to_plot:
            coefs.append(CoefficientPlottingSetting[keyword.upper()])

        return Coefficients(coefs_to_plot=coefs, cutoff=cutoff, n_largest=n_largest, name=name)

    def __init__(self, coefs_to_plot: CoefficientPlottingSettingList, cutoff: list, n_largest: list, train_dataset: Dataset = None,
                 test_dataset: Dataset = None, method: MLMethod = None, result_path: Path = None, name: str = None, hp_setting: HPSetting = None,
                 label=None, number_of_processes: int = 1):
        super().__init__(train_dataset=train_dataset, test_dataset=test_dataset, method=method, result_path=result_path,
                         name=name, hp_setting=hp_setting, label=label, number_of_processes=number_of_processes)
        self._coefs_to_plot = coefs_to_plot
        self._cutoff = cutoff
        self._n_largest = n_largest

    def _generate(self):
        PathBuilder.build(self.result_path)
        paths = []

        self._set_plotting_parameters()

        plot_data = self._retrieve_plot_data()
        plot_data["abs_coefficients"] = abs(plot_data["coefficients"])
        plot_data.sort_values(by="abs_coefficients", inplace=True, ascending=False)

        result_table_path = self._write_results_table(plot_data[["features", "coefficients"]])
        self._write_settings()

        if CoefficientPlottingSetting.ALL in self._coefs_to_plot:
            report_output_fig = self._plot(plotting_data=plot_data, output_name="all_coefficients")
            paths.append(report_output_fig)

        if CoefficientPlottingSetting.NONZERO in self._coefs_to_plot:
            nonzero_data = plot_data[plot_data["coefficients"] != 0]
            report_output_fig = self._plot(plotting_data=nonzero_data, output_name="nonzero_coefficients")
            paths.append(report_output_fig)

        if CoefficientPlottingSetting.CUTOFF in self._coefs_to_plot:
            for cutoff_val in self._cutoff:
                cutoff_data = plot_data[plot_data["abs_coefficients"] >= cutoff_val]
                report_output_fig = self._plot(plotting_data=cutoff_data, output_name="cutoff_{}_coefficients".format(cutoff_val))
                paths.append(report_output_fig)

        if CoefficientPlottingSetting.N_LARGEST in self._coefs_to_plot:
            for n_val in self._n_largest:
                n_largest_data = plot_data.nlargest(n=n_val, columns=["abs_coefficients"])
                report_output_fig = self._plot(plotting_data=n_largest_data, output_name="largest_{}_coefficients".format(n_val))
                paths.append(report_output_fig)

        return ReportResult(self.name,
                            info=f"{self._y_axis_title}s of the trained {self.method.__class__.__name__} model",
                            output_tables=[ReportOutput(result_table_path, "features and coefficients csv")],
                            output_figures=[p for p in paths if p is not None])

    def _set_plotting_parameters(self):
        if isinstance(self.method, RandomForestClassifier) or isinstance(self.method, GradientBoosting):
            self._param_field = "feature_importances"
            self._y_axis_title = "Feature importance"
        else:
            # SVM, logistic regression, ...
            self._param_field = "coefficients"
            self._y_axis_title = "Coefficient value"

    def _write_settings(self):
        if self.hp_setting is not None:
            file_path = self.result_path / "settings.yaml"
            with file_path.open("w") as file:
                yaml.dump({"preprocessing": self.hp_setting.preproc_sequence_name,
                           "encoder": self.hp_setting.encoder_name,
                           "ml_method": self.hp_setting.ml_method_name},
                          file)

    def _write_results_table(self, plotting_data):
        filepath = self.result_path / "coefficients.csv"
        plotting_data.to_csv(filepath, index=False)
        return filepath

    def _retrieve_plot_data(self):
        coefficients = self.method.get_params()[self._param_field]

        feature_names = self._retrieve_feature_names()

        return pd.DataFrame({"coefficients": coefficients, "features": feature_names})

    def _retrieve_feature_names(self):
        if self.train_dataset and self.train_dataset.encoded_data:
            return self.train_dataset.encoded_data.feature_names

    def _plot(self, plotting_data, output_name):
        if plotting_data.empty:
            logging.warning(f"Coefficients: empty data subset specified, skipping {output_name} plot...")
        else:

            filename = self.result_path / f"{output_name}.html"

            figure = px.bar(plotting_data, x='features', y='coefficients', template='plotly_white',
                            title=f"{type(self.method).__name__}{' (' + self.method.name + ') - ' if self.method.name is not None else ' - '}"
                                  f"{' '.join(output_name.split('_'))}")
            figure.update_traces(marker_color=px.colors.sequential.Teal[3])

            filename = PlotlyUtil.write_image_to_file(figure, filename, 1)

            return ReportOutput(filename)

    def check_prerequisites(self):

        run_report = True

        if not any([isinstance(self.method, legal_method) for legal_method in (RandomForestClassifier, LogisticRegression, SVM, SVC, GradientBoosting)]):
            logging.warning(f"Coefficients report can only be created for RandomForestClassifier, LogisticRegression, SVC, or SVM, but got "
                            f"{type(self.method).__name__} instead. Coefficients report will not be created.")
            run_report = False

        return run_report

    @staticmethod
    def get_documentation():
        doc = str(Coefficients.__doc__)
        valid_values = str([option.name for option in CoefficientPlottingSetting])[1:-1].replace("'", "`")
        mapping = {
            "For options see :py:obj:`~immuneML.reports.ml_reports.CoefficientPlottingSetting.CoefficientPlottingSetting`.":
                f"Valid values are: {valid_values}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



from dataclasses import dataclass
from pathlib import Path

from immuneML.data_model import bnp_util
from immuneML.data_model.SequenceParams import RegionType
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType


@dataclass
class EncoderParams:
    result_path: Path = None
    label_config: LabelConfiguration = None
    pool_size: int = 4
    model: dict = None
    learn_model: bool = True
    encode_labels: bool = True
    sequence_type: SequenceType = SequenceType.AMINO_ACID
    region_type: RegionType = RegionType.IMGT_CDR3

    def get_sequence_field_name(self):
        return bnp_util.get_sequence_field_name(self.region_type, self.sequence_type)

    def get_seq_name_for_seq_object(self):
        return 'sequence' if self.sequence_type == SequenceType.NUCLEOTIDE else "sequence_aa"



from typing import List

import numpy as np
import pandas as pd

from immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder import CompAIRRSequenceAbundanceEncoder
from immuneML.encodings.abundance_encoding.KmerAbundanceEncoder import KmerAbundanceEncoder
from immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder import SequenceAbundanceEncoder
from immuneML.hyperparameter_optimization.states.HPItem import HPItem
from immuneML.util.ParameterValidator import ParameterValidator


class SequenceAnalysisHelper:

    @staticmethod
    def compute_overlap_matrix(hp_items: List[HPItem]):
        for hp_item in hp_items:
            ParameterValidator.assert_in_valid_list(hp_item.encoder.__class__,
                                                    [SequenceAbundanceEncoder, CompAIRRSequenceAbundanceEncoder, KmerAbundanceEncoder],
                                                    'Overlap matrix computation', 'encoder')

        overlap_matrix = np.zeros((len(hp_items), len(hp_items)))

        import_sequences_as_set = lambda path: set(pd.read_csv(path).apply(frozenset, axis=1).values.tolist())

        for index1 in range(len(hp_items)):
            overlap_matrix[index1, index1] = 100
            sequences1 = import_sequences_as_set(hp_items[index1].encoder.relevant_sequence_path)
            if len(sequences1) == 0:
                return None
            for index2 in range(index1 + 1, len(hp_items)):
                sequences2 = import_sequences_as_set(hp_items[index2].encoder.relevant_sequence_path)
                if len(sequences2) == 0:
                    return None
                intersection = sequences1.intersection(sequences2)
                overlap_matrix[index1, index2] = round(len(intersection) * 100 / min(len(sequences1), len(sequences2)), 2)
                overlap_matrix[index2, index1] = overlap_matrix[index1, index2]

        return overlap_matrix



from pathlib import Path

from immuneML.util.Logger import print_log
from immuneML.util.ReflectionHandler import ReflectionHandler
from immuneML.workflows.instructions.Instruction import Instruction


class SemanticModel:

    def __init__(self, instructions: list, result_path: Path, output=None):
        assert all(isinstance(instruction, Instruction) for instruction in instructions), \
            "SemanticModel: error occurred in parsing: check instruction definitions in the configuration file."
        self.instructions = instructions
        self.result_path = result_path
        self.output = output

    def run(self):
        instruction_states = self.run_instructions()
        if self.output is not None:
            self.build_reports(instruction_states)
        return instruction_states

    def build_reports(self, instruction_states):
        report_builder = self.make_report_builder()
        print_log(f"Generating {self.output['format']} reports...", include_datetime=True)
        result_path = report_builder.build(instruction_states, self.result_path)
        print_log(f"{self.output['format']} reports are generated.", include_datetime=True)
        return result_path

    def run_instructions(self) -> list:
        instruction_states = []
        for index, instruction in enumerate(self.instructions):
            print_log(f"Instruction {index+1}/{len(self.instructions)} has started.", include_datetime=True)
            result = instruction.run(result_path=self.result_path)
            instruction_states.append(result)
            print_log(f"Instruction {index+1}/{len(self.instructions)} has finished.", include_datetime=True)
        return instruction_states

    def make_report_builder(self):
        report_builder = ReflectionHandler.get_class_by_name(f"{self.output['format']}Builder", "presentation/")
        return report_builder



from pathlib import Path
from typing import Tuple, List

import numpy as np
import plotly.express as px

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.encoding_reports.FeatureReport import FeatureReport


class FeatureDistribution(FeatureReport):
    """
    Encoding a dataset results in a numeric matrix, where the rows are examples (e.g., sequences, receptors, repertoires)
    and the columns are features. For example, when :ref:`KmerFrequency` encoder is used, the features are the
    k-mers (AAA, AAC, etc..) and the feature values are the frequencies per k-mer.

    This report plots the distribution of feature values.
    For each feature, a violin plot is created to show the distribution of feature values across all examples.
    The violin plots can be separated into different colors or facets using metadata labels
    (for example: plot the feature distributions of 'cohort1', 'cohort2' and 'cohort3' in different colors to spot biases).

    See also: :py:obj:`~immuneML.reports.encoding_reports.FeatureValueBarplot.FeatureValueBarplot` report to plot
    a simple bar chart per feature (average across examples), rather than the entire distribution.
    Or :py:obj:`~immuneML.reports.encoding_reports.FeatureDistribution.FeatureComparison` report to compare
    features across binary metadata labels (e.g., plot the feature value of 'sick' repertoires on the x axis,
    and 'healthy' repertoires on the y axis).


    Example output:

    .. image:: ../../_static/images/reports/feature_distribution.png
       :alt: Feature distribution report example
       :width: 750


    **Specification arguments:**

    - color_grouping_label (str): The label that is used to color each bar, at each level of the grouping_label.

    - row_grouping_label (str): The label that is used to group bars into different row facets.

    - column_grouping_label (str): The label that is used to group bars into different column facets.

    - mode (str): either 'normal', 'sparse' or 'auto' (default). in the 'normal' mode there are normal boxplots
      corresponding to each column of the encoded dataset matrix; in the 'sparse' mode all zero cells are eliminated before
      passing the data to the boxplots. If mode is set to 'auto', then it will automatically
      set to 'sparse' if the density of the matrix is below 0.01

    - x_title (str): x-axis label

    - y_title (str): y-axis label

    - plot_top_n (int): plot n of the largest features on average separately (useful when there are too many features
      to plot at the same time). The n features are chosen based on the average feature values across all examples
      without grouping by labels.

    - plot_bottom_n (int): plot n of the smallest features on average separately (useful when there are too many
      features to plot at the same time). The n features are chosen based on the average feature values across all
      examples without grouping by labels.

    - plot_all_features (bool): whether to plot all (might be slow for large number of features)

    - error_function (str): which error function to use for the error bar. Options are 'std' (standard deviation) or
      'sem' (standard error of the mean). Default: std.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_fdistr_report:
                    FeatureDistribution:
                        mode: sparse
                        plot_all_features: True
                        plot_top_n: 10
                        plot_bottom_n: 10

    """

    @classmethod
    def build_object(cls, **kwargs):
        return FeatureDistribution(**kwargs)

    def __init__(self, dataset: Dataset = None, result_path: Path = None, color_grouping_label: str = None,
                 row_grouping_label=None, column_grouping_label=None,
                 mode: str = 'auto', x_title: str = None, y_title: str = None, number_of_processes: int = 1,
                 name: str = None, error_function: str = None,
                 plot_top_n: int = None, plot_bottom_n: int = None, plot_all_features: bool = True):
        super().__init__(dataset=dataset, result_path=result_path, color_grouping_label=color_grouping_label,
                         row_grouping_label=row_grouping_label, column_grouping_label=column_grouping_label,
                         number_of_processes=number_of_processes, name=name, error_function=error_function)
        self.x_title = x_title if x_title is not None else self.x
        self.y_title = y_title if y_title is not None else "value"
        self.mode = mode
        self.result_name = "feature_distributions"
        self.plot_all_features = plot_all_features
        self.plot_top_n = plot_top_n
        self.plot_bottom_n = plot_bottom_n

    def _generate(self):
        result = self._generate_report_result()
        result.info = "Each boxplot represents one feature of the encoded data matrix, and shows the distribution of values for that feature."
        return result

    def _plot(self, data_long_format, mode='sparse') -> Tuple[List[ReportOutput], List[ReportOutput]]:

        plotting_data_dict = self._get_plotting_data_dict(data_long_format)

        output_figures = []
        output_tables = []
        sparse_threshold = 0.01

        for key, data in plotting_data_dict.items():

            if self.mode == 'auto':
                if (data.value == 0).mean() < sparse_threshold:
                    self.mode = 'normal'
                else:
                    self.mode = 'sparse'

            if self.mode == 'sparse':
                output_figures, output_tables = self._plot_sparse(key, data, output_tables, output_figures)
            elif self.mode == 'normal':
                output_figures, output_tables = self._plot_normal(key, data, output_tables, output_figures)

        return output_figures, output_tables

    def _plot_sparse(self, key, plotting_data, output_tables, output_figures) -> Tuple[
        List[ReportOutput], List[ReportOutput]]:
        columns_to_filter = [self.x, "value"]
        for optional_column in [self.color, self.facet_row, self.facet_column]:
            if optional_column is not None:
                columns_to_filter.append(optional_column)

        plotting_data_filtered = plotting_data.loc[plotting_data.value != 0, columns_to_filter]
        columns_to_filter.remove("value")
        total_counts = plotting_data_filtered.groupby(columns_to_filter, as_index=False).agg(
            {"value": 'sum'})
        plotting_data_filtered = plotting_data_filtered.merge(total_counts,
                                                              on=self.x,
                                                              how="left",
                                                              suffixes=('', '_sum')) \
            .fillna(0) \
            .sort_values(by=self.x) \
            .reset_index(drop=True)

        plotting_data_filtered.sort_values([self.color], ascending=[False], inplace=True)

        figure = px.violin(plotting_data_filtered, x=self.x, y="value", color=self.color,
                           facet_row=self.facet_row, facet_col=self.facet_column,
                           labels={
                               "value": self.y_title,
                               self.x: self.x_title,
                           }, template='plotly_white',
                           color_discrete_sequence=px.colors.diverging.Tealrose)
        figure.update_layout(xaxis={'categoryorder': 'total descending'})


        file_path = self.result_path / f"{self.result_name}_{key}.html"
        plotting_data.to_csv(self.result_path / f"{self.result_name}_{key}.csv", index=False)

        file_path = PlotlyUtil.write_image_to_file(figure, file_path)
        output_tables.append(
            ReportOutput(path=self.result_path / f"{self.result_name}_{key}.csv", name=f"{self.result_name} {key}"))

        output_figures.append(ReportOutput(path=file_path,
                                           name=f"Distributions of feature values ({key}, sparse data, zero values filtered)"))

        return output_figures, output_tables

    def _plot_normal(self, key, plotting_data, output_tables, output_figures) -> Tuple[
        List[ReportOutput], List[ReportOutput]]:
        figure = px.violin(plotting_data, x=self.x, y="value", color=self.color,
                           facet_row=self.facet_row, facet_col=self.facet_column,
                           labels={
                               "value": self.y_title,
                               self.x: self.x_title,
                           }, template='plotly_white',
                           color_discrete_sequence=px.colors.diverging.Tealrose)

        figure.update_layout(xaxis={'categoryorder': 'total descending'})

        file_path = self.result_path / f"{self.result_name}_{key}.html"
        plotting_data.to_csv(self.result_path / f"{self.result_name}_{key}.csv", index=False)

        file_path = PlotlyUtil.write_image_to_file(figure, file_path)
        output_tables.append(
            ReportOutput(path=self.result_path / f"{self.result_name}_{key}.csv", name=f"{self.result_name} {key}"))

        output_figures.append(ReportOutput(path=file_path,
                                           name=f"Distributions of feature values ({key})"))

        return output_figures, output_tables

    def _get_plotting_data_dict(self, data_long_format):
        plotting_data_all = data_long_format
        groupby_cols_features = [self.x]
        data_groupedby_features = self._get_grouped_data(data_long_format, groupby_cols_features)

        plotting_data_dict = {'all': plotting_data_all} if self.plot_all_features else {}

        if self.plot_top_n:
            top_n_features = data_groupedby_features.iloc[
                np.argpartition(data_groupedby_features['valuemean'].values, -self.plot_top_n)[-self.plot_top_n:]][
                self.x]
            plotting_data_dict[f'top_{self.plot_top_n}'] = plotting_data_all.loc[
                plotting_data_all[self.x].isin(top_n_features)]

        if self.plot_bottom_n:
            bottom_n_features = data_groupedby_features.iloc[
                np.argpartition(data_groupedby_features['valuemean'].values, self.plot_bottom_n)[:self.plot_bottom_n]][
                self.x]
            plotting_data_dict[f'bottom_{self.plot_bottom_n}'] = plotting_data_all.loc[
                plotting_data_all[self.x].isin(bottom_n_features)]

        return plotting_data_dict



from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Union

from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.environment.SequenceType import SequenceType
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.reports.ReportResult import ReportResult
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringItem, ClusteringSetting


@dataclass
class ClusteringConfig:
    name: str
    dataset: Dataset
    metrics: List[str]
    split_config: SplitConfig
    validation_type: List[str]
    clustering_settings: List[ClusteringSetting]
    region_type: RegionType = RegionType.IMGT_CDR3
    label_config: LabelConfiguration = None
    sequence_type: SequenceType = SequenceType.AMINO_ACID


@dataclass
class ClusteringItemResult:
    item: ClusteringItem
    report_results: List[ReportResult] = field(default_factory=list)


@dataclass
class ClusteringResultPerRun:
    run_id: int
    run_type: str
    items: Dict[str, ClusteringItemResult] = field(default_factory=dict)

    def get_cl_item(self, cl_setting: Union[str, ClusteringSetting]):
        key = cl_setting if isinstance(cl_setting, str) else cl_setting.get_key()
        return self.items[key].item


@dataclass
class ClusteringResults:
    discovery: ClusteringResultPerRun = None
    method_based_validation: ClusteringResultPerRun = None
    result_based_validation: ClusteringResultPerRun = None


@dataclass
class ClusteringState:
    name: str
    config: ClusteringConfig
    result_path: Path = None
    clustering_items: List[ClusteringResults] = field(default_factory=list)
    predictions_paths: List[Dict[str, Path]] = None
    discovery_datasets: List[Dataset] = None
    validation_datasets: List[Dataset] = None
    clustering_report_results: List[ReportResult] = field(default_factory=list)

    def add_cl_result_per_run(self, run_id: int, analysis_desc: str, cl_item_result: ClusteringResultPerRun):
        if len(self.clustering_items) <= run_id:
            self.clustering_items.append(ClusteringResults(**{analysis_desc: cl_item_result}))
        else:
            setattr(self.clustering_items[run_id], analysis_desc, cl_item_result)



import logging
import os
from pathlib import Path

from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.dsl.DefaultParamsLoader import DefaultParamsLoader
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.ImportHelper import ImportHelper
from immuneML.util.Logger import print_log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.ReflectionHandler import ReflectionHandler


class MatchedReferenceUtil:
    """
    Utility class for MatchedSequencesEncoder and MatchedReceptorsEncoder
    """

    @staticmethod
    def prepare_reference(reference_params: dict, location: str, paired: bool):
        try:
            ParameterValidator.assert_keys(list(reference_params.keys()), ["format", "params"], location, "reference")

            seq_import_params = reference_params["params"] if "params" in reference_params else {}

            assert os.path.isfile(
                seq_import_params["path"]), f"{location}: the file {seq_import_params['path']} does not exist. " \
                                            f"Specify the correct path under reference."

            if "is_repertoire" in seq_import_params:
                assert seq_import_params[
                           "is_repertoire"] is False, f"{location}: is_repertoire must be False for SequenceImport"
            else:
                seq_import_params["is_repertoire"] = False

            if "paired" in seq_import_params:
                assert seq_import_params["paired"] == paired, f"{location}: paired must be {paired} for SequenceImport"
            else:
                seq_import_params["paired"] = paired

            format_str = reference_params["format"]

            import_class = ReflectionHandler.get_class_by_name(f"{format_str}Import")
            assert import_class is not None, (f"{MatchedReferenceUtil.__name__}: {format_str} could not be imported. "
                                              f"Check if the format name has been written correctly.")
            default_params = DefaultParamsLoader.load(EnvironmentSettings.default_params_path / "datasets",
                                                      DefaultParamsLoader.convert_to_snake_case(format_str))

            params = {**default_params, **seq_import_params}

            path = Path(reference_params['params']['path'])
            params['result_path'] = PathBuilder.build(
                path.parent / 'iml_imported' if path.is_file() else path / 'iml_imported')

            if format_str == "SingleLineReceptor":
                receptors = list(import_class(params, 'tmp_receptor_dataset').import_dataset().get_data())
            else:
                receptors = list(import_class(params=params, dataset_name="tmp_dataset").import_dataset().get_data())

            assert len(
                receptors) > 0, f"MatchedReferenceUtil: The total number of imported reference {'receptors' if paired else 'sequences'} is 0, please ensure that reference import is specified correctly."

            check_imported_references(paired, receptors, seq_import_params)

            logging.info(
                f"MatchedReferenceUtil: successfully imported {len(receptors)} reference {'receptors' if paired else 'sequences'}.")

            return receptors
        except Exception as e:
            print_log(f"MatchedReferenceUtil: Error while preparing reference: {e}", logging.ERROR)
            raise e


def check_imported_references(paired, receptors, seq_import_params):
    check_for_duplicates(paired, receptors, seq_import_params)
    # check_genes(paired, receptors, seq_import_params)


def check_genes(paired, receptors, seq_import_params):
    import re

    pattern = re.compile(r'^[A-Za-z]+[0-9]+(?:-[0-9]+)?(?:\*[0-9]+)?(?:/[A-Za-z]+[0-9]+)?$')

    if not paired:
        all_v_genes = set(seq.v_call for seq in receptors if seq.v_call is not None and seq.v_call != '')
        all_j_genes = set(seq.j_call for seq in receptors if seq.j_call is not None and seq.j_call != '')
    else:
        all_v_genes = (set(receptor.chain_1.v_call for receptor in receptors if receptor.chain_1.v_call is not None and receptor.chain_1.v_call != '')
                       .union(set(receptor.chain_2.v_call for receptor in receptors if receptor.chain_2.v_call
                                  is not None and receptor.chain_2.v_call != '')))
        all_j_genes = set(receptor.chain_1.j_call for receptor in receptors if receptor.chain_1.j_call is not None and receptor.chain_1.j_call != '').union(
            set(receptor.chain_2.j_call for receptor in receptors if receptor.chain_2.j_call is not None and receptor.chain_2.j_call != ''))

    for gene_name, gene_list in [('V', all_v_genes), ('J', all_j_genes)]:
        if len(gene_list) > 0:
            assert all(pattern.match(gene) for gene in gene_list), \
                (f"{MatchedReferenceUtil.__name__}: The {gene_name} gene names in the reference sequences "
                 f"({seq_import_params['path']}) do not follow the IMGT nomenclature. Please ensure that the "
                 f"{gene_name} gene names are in the correct format (e.g., TRBV5-1*01 for V genes, TRBJ2-7*01 "
                 f"for J genes). Found {gene_name} genes: {gene_list}")


def check_for_duplicates(paired, receptors, seq_import_params):
    if not paired:
        all_sequences = [f'{seq.v_call}_{seq.get_sequence()}_{seq.j_call}' for seq in receptors]
        unique_sequences = set(all_sequences)
        if len(unique_sequences) < len(receptors):
            logging.warning(f"MatchedReferenceUtil: The reference sequences ({seq_import_params['path']}) "
                            f"contain duplicates: {len(all_sequences) - len(unique_sequences)} sequences are "
                            f"not unique. This will result in duplicate features.")



from enum import Enum


class NormalizationType(Enum):
    """
    Different normalization types for vectors.

    - RELATIVE_FREQUENCY: Each value is divided by the sum of all values (L1 normalization).
    - L2: Each value is divided by the L2 norm (Euclidean norm) [focuses on the direction of the vector, not the magnitude].
    - MAX: Each value is divided by the maximum value in the vector.
    - BINARY: Each value is set to 1 if it is greater than 0, otherwise it is set to 0.
    - NONE: No normalization is applied.

    Used in encodings like GeneFrequencyEncoder, KmerFrequencyEncoder, etc.
    """

    RELATIVE_FREQUENCY = "l1"
    L2 = "l2"
    MAX = "max"
    BINARY = "binary"
    NONE = "none"



import itertools
import os
from pathlib import Path
from typing import Union

from immuneML.workflows.instructions.apply_gen_model.ApplyGenModelInstruction import ApplyGenModelState
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.ml_methods.util.Util import Util as MLUtil
from immuneML.presentation.TemplateParser import TemplateParser
from immuneML.presentation.html.Util import Util
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.train_gen_model.TrainGenModelInstruction import TrainGenModelState


class GenModelHTMLBuilder:
    CSS_PATH = EnvironmentSettings.html_templates_path / "css/custom.css"

    @staticmethod
    def build(state: Union[ApplyGenModelState, TrainGenModelState]) -> Path:
        base_path = PathBuilder.build(state.result_path / "../HTML_output/")
        html_map = GenModelHTMLBuilder.make_html_map(state, base_path)
        result_file = base_path / f"GenModel_{state.name}.html"

        TemplateParser.parse(template_path=EnvironmentSettings.html_templates_path / "GenModel.html",
                             template_map=html_map, result_path=result_file)

        return result_file

    @staticmethod
    def make_html_map(state, base_path: Path) -> dict:
        exported_datasets = [{'name': key,
                              'path': os.path.relpath(path=Util.make_downloadable_zip(state.result_path, path),
                                                      start=base_path)}
                             for key, path in state.exported_datasets.items()]
        html_map = {
            "css_style": Util.get_css_content(GenModelHTMLBuilder.CSS_PATH),
            "name": state.name,
            'immuneML_version': MLUtil.get_immuneML_version(),
            "full_specs": Util.get_full_specs_path(base_path),
            "logfile": Util.get_logfile_path(base_path),
            "function": "Applied" if isinstance(state, ApplyGenModelState) else "Trained",
            'exported_datasets': exported_datasets,
            "show_exported_datasets": len(exported_datasets) > 0,
        }

        html_map = {**html_map, **{
            'show_reports': any(len(rep_results) > 0 for rep_results in state.report_results.values()),
            'reports': list(itertools.chain.from_iterable(
                [Util.to_dict_recursive(Util.update_report_paths(report_result, base_path), base_path)
                 for report_result in state.report_results[report_type]]
                for report_type in state.report_results.keys())),
        }}

        if hasattr(state, "generated_dataset") and state.generated_dataset is not None:
            if "generated_dataset" in state.exported_datasets.keys():
                html_map = {**html_map,
                            **Util.make_dataset_html_map(state.generated_dataset, "generated_dataset"),
                            **{"show_generated_dataset": True}}

        if hasattr(state, "combined_dataset") and state.combined_dataset is not None:
            if "combined_dataset" in state.exported_datasets.keys():
                html_map = {**html_map,
                            **Util.make_dataset_html_map(state.combined_dataset, "combined_dataset"),
                            **{"show_combined_dataset": True}}

        return html_map



from pathlib import Path

from immuneML.api.galaxy.GalaxyTool import GalaxyTool
from immuneML.api.galaxy.build_train_ml_model_yaml import main
from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.PathBuilder import PathBuilder


class RepertoireClassificationTool(GalaxyTool):
    def __init__(self, args, result_path: Path):
        self.args = args
        super().__init__(None, result_path)

    def _run(self):
        yaml_path = main(self.args)

        PathBuilder.build(self.result_path)

        app = ImmuneMLApp(yaml_path, self.result_path)
        output_file_path = app.run()

        return output_file_path



import copy
import logging
from dataclasses import field, dataclass
from pathlib import Path
from typing import Dict, List

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.data_model.bnp_util import merge_dataclass_objects, bnp_write_to_file, write_dataset_yaml
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.environment.SequenceType import SequenceType
from immuneML.hyperparameter_optimization.config.ManualSplitConfig import ManualSplitConfig
from immuneML.hyperparameter_optimization.config.SplitConfig import SplitConfig
from immuneML.hyperparameter_optimization.config.SplitType import SplitType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.reports.train_gen_model_reports.TrainGenModelReport import TrainGenModelReport
from immuneML.reports.gen_model_reports.GenModelReport import GenModelReport
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.steps.data_splitter.DataSplitter import DataSplitter
from immuneML.workflows.steps.data_splitter.DataSplitterParams import DataSplitterParams


@dataclass
class TrainGenModelState:
    result_path: Path = None
    name: str = None
    gen_examples_count: int = None
    model_paths: Dict[str, Path] = field(default_factory=dict)
    generated_dataset: Dataset = None
    exported_datasets: Dict[str, Path] = field(default_factory=dict)
    report_results: Dict[str, List[ReportResult]] = field(
        default_factory=lambda: {'data_reports': [], 'ml_reports': [], 'gen_ml_reports': [], 'instruction_reports': []})
    combined_dataset: Dataset = None
    train_dataset: Dataset = None
    test_dataset: Dataset = None
    training_percentage: float = None
    split_strategy: SplitType = None
    split_config: ManualSplitConfig = None


class TrainGenModelInstruction(Instruction):
    """

    TrainGenModel instruction implements training generative AIRR models on receptor level. Models that can be trained
    for sequence generation are listed under Generative Models section.

    This instruction takes a dataset as input which will be used to train a model, the model itself, and the number of
    sequences to generate to illustrate the applicability of the model. It can also produce reports of the fitted model
    and reports of original and generated sequences.

    To use the generative model previously trained with immuneML, see :ref:`ApplyGenModel` instruction.


    **Specification arguments:**

    - dataset: dataset to use for fitting the generative model; it has to be defined under definitions/datasets

    - methods: which methods to fit (defined previously under definitions/ml_methods); for compatibility with previous
      versions 'method' with a single method can also be used, but the single method option will be removed in the
      future.

    - number_of_processes (int): how many processes to use for fitting the model

    - gen_examples_count (int): how many examples (sequences, repertoires) to generate from the fitted model

    - reports (list): list of report ids (defined under definitions/reports) to apply after fitting a generative model
      and generating gen_examples_count examples; these can be data reports (to be run on generated examples), ML
      reports (to be run on the fitted model)

    - split_strategy (str): strategy to use for splitting the dataset into training and test datasets; valid options are
      RANDOM and MANUAL (in which case train and test set are fixed); default is RANDOM

    - training_percentage (float): percentage of the dataset to use for training the generative model if split_strategy
      parameter is RANDOM. If set to 1, the
      full dataset will be used for training and the test dataset will be the same as the training dataset. Default
      value is 0.7. When export_combined_dataset is set to True, the splitting of sequences into train, test, and
      generated will be shown in column dataset_split.

    - manual_config (dict): if split_strategy is set to MANUAL, this parameter can be used to specify the ids of examples
      that should be in train and test sets; the paths to csv files with ids for train and test data should be provided
      under keys 'train_metadata_path' and 'test_metadata_path'

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_train_gen_model_inst: # user-defined instruction name
                type: TrainGenModel
                dataset: d1 # defined previously under definitions/datasets
                methods: [model1] # defined previously under definitions/ml_methods
                gen_examples_count: 100
                number_of_processes: 4
                training_percentage: 0.7
                split_strategy: RANDOM # optional, default is RANDOM
                export_generated_dataset: True
                export_combined_dataset: False
                reports: [data_rep1, ml_rep2]

            my_train_gen_model_with_manual_split: # another instruction example
                type: TrainGenModel
                dataset: d1 # defined previously under definitions/datasets
                methods: [m1, m2]
                gen_examples_count: 100
                split_strategy: MANUAL
                split_config:
                    train_metadata_path: path/to/train_metadata.csv # path to csv file with ids of examples in train set
                    test_metadata_path: path/to/test_metadata.csv # path to csv file with ids of examples in test set
                export_generated_dataset: True
                export_combined_dataset: False
                reports: [data_rep1, ml_rep2]

    """

    MAX_ELEMENT_COUNT_TO_SHOW = 10

    def __init__(self, dataset: Dataset = None, methods: List[GenerativeModel] = None,
                 number_of_processes: int = 1, gen_examples_count: int = 100, result_path: Path = None,
                 name: str = None, reports: list = None, export_generated_dataset: bool = True,
                 export_combined_dataset: bool = False, training_percentage: float = None,
                 split_strategy: SplitType = SplitType.RANDOM, split_config: ManualSplitConfig = None):
        self.state = TrainGenModelState(result_path, name, gen_examples_count, training_percentage=training_percentage,
                                        split_config= split_config, split_strategy=split_strategy)
        self.methods = methods
        self.reports = reports
        self.dataset = dataset
        self.number_of_processes = number_of_processes
        self.export_generated_dataset = export_generated_dataset
        self.export_combined_dataset = export_combined_dataset
        self.generated_datasets = {}

    def run(self, result_path: Path) -> TrainGenModelState:
        self._set_path(result_path)
        self._split_dataset()
        self._fit_model()
        self._save_model()
        self._gen_data()
        if self.export_generated_dataset:
            self._export_generated_dataset()
        self._make_and_export_combined_dataset()
        self._run_reports()

        return self.state

    def _split_dataset(self):
        if self.state.split_strategy == SplitType.RANDOM:
            if self.state.training_percentage != 1:
                split_params = DataSplitterParams(dataset=self.dataset, split_strategy=SplitType.RANDOM, split_count=1,
                                                  training_percentage=self.state.training_percentage,
                                                  paths=[self.state.result_path])
                train_datasets, test_datasets = DataSplitter.run(split_params)
                self.state.train_dataset = train_datasets[0]
                self.state.test_dataset = test_datasets[0]
            else:
                logging.info(f"{TrainGenModelInstruction.__name__}: training_percentage was set to 1 meaning that the full "
                             f"dataset will be used for fitting the generative model. All resulting comparison reports "
                             f"will then use the full original dataset as opposed to independent test dataset if the "
                             f"training percentage was less than 1.")
                self.state.train_dataset = self.dataset
                self.state.test_dataset = self.dataset
        elif self.state.split_strategy == SplitType.MANUAL:
            split_params = DataSplitterParams(self.dataset, self.state.split_strategy, 1,
                                              paths=[self.state.result_path],
                                              split_config=SplitConfig(self.state.split_strategy, split_count=1,
                                                                       manual_config=self.state.split_config))
            train_datasets, test_datasets = DataSplitter.manual_split(split_params)
            self.state.train_dataset = train_datasets[0]
            self.state.test_dataset = test_datasets[0]
        else:
            raise ValueError(f"{TrainGenModelInstruction.__name__}: {self.state.name}: "
                             f"split_strategy {self.state.split_strategy} is not supported for TrainGenModel instruction.")

    def _fit_model(self):
        for ind, method in enumerate(self.methods):
            print_log(f"{self.state.name}: starting to fit the model {method.name} ({ind + 1}/{len(self.methods)})",
                      True)
            method.fit(self.state.train_dataset, self.state.result_path)
            print_log(f"{self.state.name}: fitted the model {method.name} ({ind + 1}/{len(self.methods)})", True)

    def _make_combined_dataset(self):
        path = PathBuilder.build(self.state.result_path / 'combined_dataset')

        gen_data_list = []
        data_origin_list = []

        for model_name, gen_dataset in self.generated_datasets.items():
            gen_data = self._get_dataclass_object_from_dataset(gen_dataset,
                                                               [model_name] * self.state.gen_examples_count)
            gen_data_list.append(gen_data)
            data_origin_list.append(model_name)

        if self.state.training_percentage < 1:
            data_origin_list.extend(['original_train', 'original_test'])

            org_data = self._get_dataclass_object_from_dataset(self.state.train_dataset,
                                                               ['original_train'] * self.state.train_dataset.get_example_count())
            test_data = self._get_dataclass_object_from_dataset(self.state.test_dataset,
                                                                ['original_test'] * self.state.test_dataset.get_example_count())

            combined_data = merge_dataclass_objects([org_data, test_data] + gen_data_list, fill_unmatched=True)
        else:
            data_origin_list.extend(['original_train'])
            org_data = self._get_dataclass_object_from_dataset(self.dataset,
                                                               ['original_train'] * self.dataset.get_example_count())
            combined_data = merge_dataclass_objects([org_data] + gen_data_list, fill_unmatched=True)

        labels = {'gen_model_name': [''] + [method.name for method in self.methods],
                  "from_gen_model": [True, False], 'data_origin': data_origin_list}

        self.state.combined_dataset = SequenceDataset.build_from_dataclass_object(combined_data, path=path,
                                                                                  name=f'combined_{self.state.name}_dataset',
                                                                                  labels=labels)

    def _get_dataclass_object_from_dataset(self, dataset: Dataset, data_origin: list):
        return dataset.data.add_fields({'data_origin': data_origin}, {'data_origin': str})

    def _make_and_export_combined_dataset(self):
        if self.export_combined_dataset and isinstance(self.dataset, SequenceDataset):
            self._make_combined_dataset()
            export_path = PathBuilder.build(self.state.result_path / 'exported_combined_dataset')
            try:
                AIRRExporter.export(self.state.combined_dataset, export_path)
                self.state.exported_datasets['combined_dataset'] = export_path
            except AssertionError as e:
                logging.warning(f"{TrainGenModelInstruction.__name__}: {self.state.name}: combined dataset could not "
                                f"be exported due to the following error: {e}.")
        elif self.export_combined_dataset:
            logging.warning(f"{TrainGenModelInstruction.__name__}: {self.state.name}: export_combined_dataset is only "
                            f"supported for sequence datasets at this point.")

    def _run_reports(self):
        self._run_reports_main()

        report_path = self._get_reports_path()
        for report in self.reports:
            original_dataset = self.state.train_dataset if self.state.training_percentage != 1 else self.dataset
            report.result_path = report_path
            if isinstance(report, TrainGenModelReport):
                for method in self.methods:
                    report.generated_dataset = self.generated_datasets[method.name]
                    report.original_dataset = original_dataset
                    report.model = method
                    self.state.report_results['instruction_reports'].append(report.generate_report())
            elif isinstance(report, DataReport):
                rep = copy.deepcopy(report)
                rep.result_path = PathBuilder.build(rep.result_path.parent / f"{rep.result_path.name}_original_dataset")
                rep.dataset = original_dataset
                rep.name = rep.name + " (original dataset)"
                self.state.report_results['data_reports'].append(rep.generate_report())
            elif isinstance(report, GenModelReport):
                for method in self.methods:
                    rep = copy.deepcopy(report)
                    rep.result_path = PathBuilder.build(rep.result_path.parent / f"{rep.result_path.name}_original_dataset")
                    rep.dataset = original_dataset
                    rep.model = method
                    rep.name = rep.name + " (original dataset)"
                    self.state.report_results['gen_ml_reports'].append(rep.generate_report())

        self._print_report_summary_log()

    def _save_model(self):
        for method in self.methods:
            self.state.model_paths[method.name] = method.save_model(
                self.state.result_path / f'trained_model_{method.name}/')

    def _gen_data(self):
        for method in self.methods:
            dataset = method.generate_sequences(self.state.gen_examples_count, 1,
                                                self.state.result_path / 'generated_sequences' / method.name,
                                                SequenceType.AMINO_ACID, False)

            self.generated_datasets[method.name] = dataset
            print_log(f"{self.state.name}: generated {self.state.gen_examples_count} examples from the fitted model",
                      True)

        self.state.generated_dataset = self.merge_datasets(list(self.generated_datasets.values()),
                                                           self.state.result_path / 'combined_generated_dataset')

    def merge_datasets(self, datasets: List[SequenceDataset], result_path: Path) -> SequenceDataset:
        if len(datasets) == 1:
            return datasets[0]
        else:
            merged_data = merge_dataclass_objects([dataset.data for dataset in datasets])
            merged_dataset = SequenceDataset.build_from_dataclass_object(merged_data, PathBuilder.build(result_path))
            return merged_dataset

    def _export_generated_dataset(self):
        AIRRExporter.export(self.state.generated_dataset, self.state.result_path / f'exported_gen_dataset')
        self.state.exported_datasets['generated_dataset'] = self.state.result_path / 'exported_gen_dataset'

    def _run_reports_main(self):
        report_path = self._get_reports_path()
        for report in self.reports:
            report.result_path = report_path
            if isinstance(report, DataReport):
                for method in self.methods:
                    rep = copy.deepcopy(report)
                    rep.dataset = self.generated_datasets[method.name]
                    rep.result_path = PathBuilder.build(
                        rep.result_path.parent / f"{rep.result_path.name}_{method.name}")
                    rep.name = rep.name + " (generated dataset from " + method.name + ")"
                    self.state.report_results['data_reports'].append(rep.generate_report())
            elif isinstance(report, MLReport):
                for method in self.methods:
                    rep = copy.deepcopy(report)
                    rep.method = method
                    rep.name = rep.name + " (from " + method.name + ")"
                    self.state.report_results['ml_reports'].append(rep.generate_report())

    def _print_report_summary_log(self):
        if len(self.reports) > 0:
            gen_rep_count = len(self.state.report_results['ml_reports']) + len(
                self.state.report_results['data_reports']) + len(self.state.report_results['gen_ml_reports'])
            print_log(f"{self.state.name}: generated {gen_rep_count} reports.", True)

    def _get_reports_path(self) -> Path:
        return PathBuilder.build(self.state.result_path / 'reports')

    def _set_path(self, result_path):
        self.state.result_path = PathBuilder.build(result_path / self.state.name)



import logging
from pathlib import Path

import pandas as pd
import plotly.express as px

from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder


class RepertoireClonotypeSummary(DataReport):
    """
    Shows the number of distinct clonotypes per repertoire in a given dataset as a bar plot.

    **Specification arguments:**

    - color_label (str): the label to color the bar plot by (optional, default: None)

    - facet_label (str): the label to facet the bar plot by (optional, default: None)

    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_clonotype_summary_rep:
                    RepertoireClonotypeSummary:
                        color_label: celiac
                        facet_label: hla


    """

    def __init__(self, dataset: Dataset = None, result_path: Path = None, name: str = None, number_of_processes: int = 1,
                 color_label: str = None, facet_label: str = None):
        super().__init__(dataset, result_path, name, number_of_processes)
        self.color_label = color_label
        self.facet_label = facet_label

    @classmethod
    def build_object(cls, **kwargs):
        return RepertoireClonotypeSummary(**kwargs)

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        return self._safe_plot()

    def add_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        column_names = []
        if self.color_label:
            column_names.append(self.color_label)
        if self.facet_label:
            column_names.append(self.facet_label)

        if len(column_names) > 0:
            metadata = self.dataset.get_metadata(column_names, return_df=True)
            return pd.concat([df, metadata], axis=1)
        else:
            return df

    def _plot(self) -> ReportResult:
        clonotypes = pd.DataFrame({'clonotype_count': [self._get_clonotype_count(repertoire)
                                   for repertoire in self.dataset.get_data()]})

        clonotypes['repertoire_id'] = self.dataset.get_example_ids()
        clonotypes = self.add_labels(clonotypes)
        clonotypes.sort_values(by='clonotype_count', ascending=False, inplace=True)
        clonotypes['repertoire_index'] = clonotypes.groupby(self.facet_label).cumcount() if self.facet_label else list(range(clonotypes.shape[0]))

        fig = px.bar(clonotypes, x='repertoire_index', y='clonotype_count', facet_row=self.facet_label,
                     color=self.color_label, title='Clonotype count per repertoire',
                     color_discrete_sequence=px.colors.qualitative.Vivid)
        fig.update_layout(template="plotly_white", yaxis_title='clonotype count',
                          xaxis_title='repertoires')

        if self.facet_label:
            facet_label_counts = {str(k): v for k, v in clonotypes[self.facet_label].value_counts().to_dict().items()}
            for annotation in fig.layout.annotations:
                group_label = annotation.text
                if '=' in group_label:
                    group = group_label.split('=')[1]
                    count = facet_label_counts.get(group, 0)
                    annotation.text = f"{group_label}<br>(n={count})"

        fig.add_annotation(
            text="clonotype counts",  # Your y-axis label
            xref="paper", yref="paper",  # Use paper coordinates (0–1)
            x=-0.07, y=0.5,  # Position to the left and centered vertically
            showarrow=False,
            textangle=-90,  # Vertical orientation
            font=dict(size=14)
        )

        fig.update_layout(margin=dict(l=80))
        fig.update_yaxes(title='')

        clonotypes.to_csv(self.result_path / 'clonotype_count_per_repertoire.csv', index=False)
        fig_path = PlotlyUtil.write_image_to_file(fig, self.result_path / 'clonotype_count_per_repertoire.html',
                                       self.dataset.get_example_count())

        return ReportResult(name=self.name, info="Clonotype count per repertoire",
                            output_figures=[ReportOutput(fig_path, name='Clonotype count per repertoire')],
                            output_tables=[ReportOutput(self.result_path / 'clonotype_count_per_repertoire.csv',
                                                        name='Clonotype count per repertoire')])

    def _get_clonotype_count(self, repertoire: Repertoire) -> int:

        sequences = repertoire.data.topandas()

        sequence_count = sequences.shape[0]
        unique_sequence_count = len(sequences.groupby(['cdr3_aa', 'v_call', 'j_call']).size().reset_index(name='count'))
        if sequence_count != unique_sequence_count:
            logging.warning(f"{RepertoireClonotypeSummary.__name__}: {self.name}: for repertoire {repertoire.identifier}, "
                            f"there are {sequence_count} sequences, but {unique_sequence_count} unique (CDR3 amino acid"
                            f" sequence, V call, J call) combinations.")

        return unique_sequence_count

    def check_prerequisites(self) -> bool:
        if isinstance(self.dataset, RepertoireDataset):
            return True
        else:
            logging.warning(f"{RepertoireClonotypeSummary.__name__}: report can be generated only from "
                            f"RepertoireDataset. Skipping this report...")
            return False



import logging
from pathlib import Path
from typing import Tuple

import pandas as pd
import plotly.express as px

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.diversity_encoding.ShannonDiversityEncoder import ShannonDiversityEncoder
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder


class ShannonDiversityOverview(DataReport):
    """
    Computes Shannon diversity for each repertoire using Shannon diversity encoder and plots the
    results in a histogram, optionally stratified by labels.

    **Dataset type:**

    - Repertoire Dataset

    **Specification arguments:**

    - color_label (str): The label used to color the histogram bars. Default is None.

    - facet_row_label (str): The label used to facet the histogram into multiple rows.
      Default is None, meaning no row faceting.

    - facet_col_label (str): The label used to facet the histogram into multiple columns.
      Default is None, meaning no column faceting.

     **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                shannon_div_rep:
                    ShannonDiversityOverview:
                        color_label: disease_status


    """

    def __init__(self, dataset: RepertoireDataset = None, result_path: Path = None, name: str = None,
                 number_of_processes: int = 1, color_label: str = None, facet_row_label: str = None,
                 facet_col_label: str = None):
        super().__init__(dataset, result_path, name, number_of_processes)
        self.color_label = color_label
        self.facet_row_label = facet_row_label
        self.facet_col_label = facet_col_label

    @classmethod
    def build_object(cls, **kwargs):
        return ShannonDiversityOverview(**kwargs)

    def check_prerequisites(self) -> bool:
        valid = isinstance(self.dataset, RepertoireDataset)

        if not valid:
            logging.warning(f"ShannonDiversityOverview: Dataset must be of type RepertoireDataset, "
                            f"but got {type(self.dataset)}.")

        return valid

    def _generate(self) -> ReportResult:
        encoded_dataset = (ShannonDiversityEncoder.build_object(self.dataset)
                           .encode(self.dataset, EncoderParams(self.result_path, encode_labels=False)))

        PathBuilder.build(self.result_path)

        df, table_output = self.prepare_data(encoded_dataset)

        figure_output = self._safe_plot(encoded_df=df)

        return ReportResult(name=self.name, info="Shannon diversity per repertoire",
                            output_figures=[figure_output],
                            output_tables=[table_output])

    def prepare_data(self, encoded_dataset) -> Tuple[pd.DataFrame, ReportOutput]:
        labels = ['subject_id'] if 'subject_id' in self.dataset.labels.keys() else []
        for label in [self.color_label, self.facet_row_label, self.facet_col_label]:
            if label is not None:
                labels.append(label)

        df = pd.DataFrame({'shannon_diversity': encoded_dataset.encoded_data.examples.flatten(),
                           'repertoire_id': encoded_dataset.get_example_ids(),
                           **self.dataset.get_metadata(labels)})

        df.sort_values(by='shannon_diversity', ascending=False, inplace=True)

        df.to_csv(self.result_path / 'shannon_diversity.csv', index=False)

        return df, ReportOutput(self.result_path / 'shannon_diversity.csv', name='Shannon diversity')

    def _plot(self, encoded_df) -> ReportOutput:

        facet_labels = []
        if self.facet_row_label:
            facet_labels.append(self.facet_row_label)
        if self.facet_col_label:
            facet_labels.append(self.facet_col_label)

        encoded_df['repertoire_index'] = encoded_df.groupby(facet_labels).cumcount() \
            if len(facet_labels) > 0 else list(range(encoded_df.shape[0]))

        hover_data_cols = ['repertoire_id'] + (['subject_id'] if 'subject_id' in encoded_df.columns else [])

        fig = px.bar(encoded_df, x='repertoire_index', y='shannon_diversity', facet_row=self.facet_row_label,
                     color=self.color_label, title='Shannon diversity per repertoire', facet_col=self.facet_col_label,
                     color_discrete_sequence=px.colors.qualitative.Vivid, hover_data=hover_data_cols)
        fig.update_layout(template="plotly_white", yaxis_title='Shannon diversity',
                          xaxis_title='Repertoires sorted by Shannon diversity')

        fig.update_traces(
            hovertemplate=(
                "Repertoire id: %{customdata[0]}<br>" +
                "Subject id: %{customdata[1]}<br>" if 'subject_id' in encoded_df.columns else "" +
                "Shannon diversity: %{y}<extra></extra>"
            )
        )

        figure_path = PlotlyUtil.write_image_to_file(fig, self.result_path / 'shannon_diversity.html')

        return ReportOutput(figure_path, name='Shannon diversity')



import copy
import yaml
import numpy as np
from pathlib import Path

from immuneML.data_model.EncodedData import EncodedData
from immuneML.environment.Label import Label
from immuneML.util.PathBuilder import PathBuilder
from immuneML.ml_methods.classifiers.MLMethod import MLMethod


class SillyClassifier(MLMethod):
    """
    This SillyClassifier is a placeholder for a real ML method.
    It generates random predictions ignoring the input features.

    **Specification arguments:**

    - random_seed (int): The random seed for generating random predictions.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_silly_method:
                    SillyClassifier:
                        random_seed: 100

    """
    def __init__(self, random_seed: int = None):
        super().__init__()
        self.random_seed = random_seed
        self.silly_model_fitted = False

    def _fit(self, encoded_data: EncodedData, cores_for_training: int = 2):
        # Since the silly classifier makes random predictions and ignores training data,
        # no model is fitted during training.
        # For any other method, model fitting should be implemented here.
        self.silly_model_fitted = True

    def _predict_proba(self, encoded_data: EncodedData):
        np.random.seed(self.random_seed)

        # Generate an array containing a random prediction probability for each example
        pred_probabilities = np.random.rand(len(encoded_data.examples))

        return {self.label.name: {self.label.positive_class: pred_probabilities,
                                  self.label.get_binary_negative_class(): 1 - pred_probabilities}}

    def _predict(self, encoded_data: EncodedData):
        predictions_proba = self.predict_proba(encoded_data, self.label)
        proba_positive_class = predictions_proba[self.label.name][self.label.positive_class]

        predictions = []

        for proba in proba_positive_class:
            if proba > 0.5:
                predictions.append(self.label.positive_class)
            else:
                predictions.append(self.label.get_binary_negative_class())

        # Shorter alternative using class mapping:
        # return {self.label.name: np.array([self.class_mapping[val] for val in (proba_positive_class > 0.5).tolist()])}

        return {self.label.name: np.array(predictions)}

    def can_predict_proba(self) -> bool:
        return True

    def can_fit_with_example_weights(self) -> bool:
        return False

    def get_compatible_encoders(self):
        # Every encoder that is compatible with the ML method should be listed here.
        # The SillyClassifier can in principle be used with any encoder, few examples are listed
        from immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder import SequenceAbundanceEncoder
        from immuneML.encodings.abundance_encoding.KmerAbundanceEncoder import KmerAbundanceEncoder
        from immuneML.encodings.atchley_kmer_encoding.AtchleyKmerEncoder import AtchleyKmerEncoder
        from immuneML.encodings.distance_encoding.DistanceEncoder import DistanceEncoder
        from immuneML.encodings.diversity_encoding.EvennessProfileEncoder import EvennessProfileEncoder
        from immuneML.encodings.kmer_frequency.KmerFrequencyEncoder import KmerFrequencyEncoder
        from immuneML.encodings.motif_encoding.MotifEncoder import MotifEncoder
        from immuneML.encodings.onehot.OneHotEncoder import OneHotEncoder

        return [SequenceAbundanceEncoder, KmerAbundanceEncoder, DistanceEncoder, EvennessProfileEncoder,
                AtchleyKmerEncoder, KmerFrequencyEncoder, MotifEncoder, OneHotEncoder]

    def store(self, path: Path):
        # The most basic way of storing a model is to get the parameters in a yaml-friendly format (get_params)
        # and store this in a file.
        # Depending on the method, more files (e.g., internal pickle, pytorch or keras files) may need to be stored.
        # The 'store' method should be compatible with 'load'
        PathBuilder.build(path)
        class_parameters = self.get_params()
        params_path = path / "custom_params.yaml"

        with params_path.open('w') as file:
            yaml.dump(class_parameters, file)

    def get_params(self, for_refitting=False) -> dict:
        # Returns a yaml-friendly dictionary (only simple types, no objects) with all parameters of this ML method
        params = copy.deepcopy(vars(self))

        if self.label and not for_refitting:
            # the 'Label' object must be converted to a yaml-friendly version
            params["label"] = self.label.get_desc_for_storage()

        return params

    def load(self, path: Path):
        # The 'load' method is called on a new (untrained) object of the specific MLMethod class.
        # This method is used to load parameters from a previously trained model from storage,
        # thus creating a copy of the original trained model

        # Load the dictionary of parameters from the YAML file
        params_path = path / "custom_params.yaml"

        with params_path.open("r") as file:
            custom_params = yaml.load(file, Loader=yaml.SafeLoader)

        # Loop through the dictionary and set each parameter
        for param, value in custom_params.items():
            if hasattr(self, param):
                if param == "label":
                    # Special case: if the parameter is 'label', convert to a Label object
                    setattr(self, "label", Label(**value))
                else:
                    # Other cases: directly set the parameter to the given value
                    setattr(self, param, value)
